# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""References and stores."""

import abc
import base64
import contextlib
import json
import os
import pickle

from typing import (
  Any, Dict, Generic, Iterable, Iterator, List, Mapping, NamedTuple, Optional,
  Set, Tuple, Type, TypeVar, Union, cast)

from enact import contexts
from enact import digests
from enact import interfaces
from enact import resource_registry
from enact import serialization
from enact import types
from enact import utils



class RefError(Exception):
  """Superclass for reference related errors."""


R = TypeVar('R')
ResourceT = TypeVar('ResourceT', bound=interfaces.ResourceBase)
RefT = TypeVar('RefT', bound='Ref')


class PackedResource(NamedTuple):
  """A resource packed for storage with a corresponding reference."""
  # A resource packed for storage. This may be just the resource itself or some
  # other representation that encodes it, e.g., compressed or encrypted.
  data: interfaces.ResourceDict
  # The reference to this object as a resource dict.
  ref_dict: interfaces.ResourceDict
  # The outgoing reference ids as a set of strings.
  links: Set[str]

  def unpack(self) -> Any:
    """Unpacks the resource."""
    return self.ref().unpack(self)

  def ref(self) -> 'Ref':
    """Returns the reference to this resource."""
    r = resource_registry.from_resource_dict(self.ref_dict)
    if not isinstance(r, Ref):
      raise RefError(f'Expected a Ref, got {r}.')
    return r


def checkout(ref: 'Ref[R]') -> R:
  """Gets the reference or asserts false if None."""
  return ref.checkout()


def commit(resource: R) -> 'Ref[R]':
  """Commits a value to the store and returns a reference."""
  return Store.current().commit(resource)


@resource_registry.register
class Ref(Generic[R], interfaces.ResourceBase):
  """Represents a reference to a resource or wrappable python object.

  References are JSON encodable objects and their key-sorted json encoding
  serves as a string key for the underlying resource.

  Ref subclasses may pack resources before storage and unpack them when they are
  retrieved. This allows implementation of subtypes of Ref that perform
  end-to-end encryption or compression.
  """

  def __init__(self, digest: str, value: Optional[R]=None):
    """Initializes the reference from a digest and optionally the resource."""
    assert isinstance(digest, str), (
      'Must instantiate Ref with a string digest.')
    self._digest = digest
    self._cached: Optional[R] = value

  def _clear_cache(self):
    """Clear the cache."""
    self._cached = None

  def _set_cache(self, value: R):
    """Set the cache."""
    self._cached = resource_registry.unwrap(value)

  @property
  def digest(self) -> str:
    """Returns a unique, deterministic digest of the referenced resource."""
    return self._digest

  @property
  def id(self) -> str:  # pylint: disable=invalid-name
    """Returns a string version of this reference."""
    return json.dumps(dict(self.field_items()), sort_keys=True)

  def __hash__(self) -> int:
    """Hash representation."""
    return hash(self.id)

  @classmethod
  def from_id(cls: Type[RefT], ref_id: str) -> RefT:
    """Returns a reference from a reference ID."""
    try:
      return cast(RefT, resource_registry.from_resource_dict(
        interfaces.ResourceDict(cls, **json.loads(ref_id))))
    except json.JSONDecodeError as error:
      raise RefError(f'Invalid ref id: {ref_id}') from error

  @contextlib.contextmanager
  def modify(self) -> Iterator[R]:
    """Context manager for modifying the resource."""
    resource = self.checkout()
    yield resource
    commit(resource)
    self.set(resource)

  def checkout(self) -> R:
    """Fetches the resource from the cache or active store."""
    if (self._cached is None or
        self.from_resource(resource_registry.wrap(self._cached)) != self):
      self._set_cache(Store.current().checkout(self))
    return cast(R, self._cached)

  def __call__(self) -> R:
    """Alias for get."""
    return self.checkout()

  def set(self, resource: R):
    """Sets the reference to point to the given resource."""
    self._digest = digests.digest(resource_registry.wrap(resource))
    self._set_cache(resource)

  @classmethod
  def from_resource(cls: Type[RefT], resource: interfaces.ResourceBase) -> RefT:
    """Constructs a reference to the resource."""
    return cls(
      digest=digests.digest(resource),
      value=resource_registry.unwrap(resource))

  @classmethod
  def from_resource_dict(
      cls: Type[RefT], resource_dict: interfaces.ResourceDict) -> RefT:
    """Constructs a reference to the resource represented by resource_dict."""
    return cls(digest=digests.digest(resource_dict))

  def __eq__(self, other: Any):
    """Returns true if the other object is the same reference."""
    if not isinstance(other, Ref):  # pylint: disable=unidiomatic-typecheck
      return False
    return self.digest == other.digest

  def verify(self, resource_dict: interfaces.ResourceDict):
    """Check that this reference matches the resource dict.

    Args:
      resource_dict: The resource dict to verify.

    Raises:
      RefError: If integrity check fails.
    """
    if self != self.from_resource_dict(resource_dict):
      raise RefError(
        f'Reference {self} does not match resource dict {resource_dict}.')

  def unpack(self: 'Ref[R]', packed_resource: PackedResource) -> R:
    """Unpacks the referenced resource."""
    self.verify(packed_resource.data)
    return cast(
      R,
      resource_registry.unwrap(
        resource_registry.from_resource_dict(packed_resource.data)))

  @classmethod
  def pack(cls: Type[RefT], resource: ResourceT) -> Tuple[RefT, PackedResource]:
    """Wraps and packs the resource."""
    # TODO: This should be improved since it walks the resource twice:
    #   1. Once for walk_resource
    #   2. Once for to_resource_dict.
    # This will involve wrapping values twice when type_wrappers are involved.
    # Maybe add a visitor interface to 'to_resource_dict' function, instead of
    # using walk_resource.
    links = {
      l.id for l in utils.walk_resource(resource) if isinstance(l, Ref)}
    ref = cls.from_resource(resource)
    return ref, PackedResource(
      resource.to_resource_dict(),
      ref_dict=ref.to_resource_dict(),
      links=links)

  @classmethod
  def field_names(cls) -> Iterable[str]:
    """Returns the names of the fields of the resource."""
    yield 'digest'

  def field_values(self) -> Iterable[interfaces.FieldValue]:
    """Return a list of field values, aligned with field_names."""
    yield self._digest

  @classmethod
  def from_fields(cls: Type[RefT],
                  field_dict: Mapping[str, interfaces.FieldValue]) -> RefT:
    """Constructs the resource from a value dictionary."""
    return cls(**field_dict)  # type: ignore

  def __repr__(self) -> str:
    return f'<{type(self).__name__}: {self.digest}>'

  def set_from(self, other: Any):
    """Sets the fields of this resource from another resource."""
    if type(other) != type(self):  # pylint: disable=unidiomatic-typecheck
      raise ValueError(f'Cannot set {self} from {other}: types do not match.')
    self._digest = other._digest  # pylint: disable=protected-access
    self._cached = other._cached  # pylint: disable=protected-access


class StorageBackend(abc.ABC):
  """A storage backend."""

  @abc.abstractmethod
  def register_type(self,
                    type_key: types.TypeKey,
                    attributes: Dict[str, Optional[types.TypeDescriptor]]):
    """Register a new type to allow resource to be committed."""

  @abc.abstractmethod
  def get_type(self, type_key: types.TypeKey) -> (
      Optional[Dict[str, Optional[types.TypeDescriptor]]]):
    """Returns the type, if known."""

  @abc.abstractmethod
  def commit(self, ref_id: str, packed_resource: PackedResource):
    """Stores a packed resource."""

  @abc.abstractmethod
  def has(self, ref_ids: Iterable[str]) -> List[bool]:
    """Returns whether the storage backend has the resource."""

  @abc.abstractmethod
  def checkout(
      self, ref_ids: Iterable[str]) -> (
        List[Optional[PackedResource]]):
    """Returns a dictionary of resources or None if not available.

    Args:
      ref_ids: The reference IDs to retrieve.

    Returns:
      A list of packed resources or None if the resource is not available.
      Returned in the order of the ref_ids argument.
    """

  def get_type_keys(
      self, ref_ids: Iterable[str]) -> List[
        Optional[Set[types.TypeKey]]]:
    """Returns the types required to unpack the resources.

    The default implementation will load all resource data and extract only
    the type info. This should be overridden if more efficiency is required.

    Args:
      ref_ids: The reference IDs to retrieve.

    Returns:
      A list of sets of types or None, in the order of the ref_ids argument.
      None is returned if a reference cannot be resolved.
    """
    result: List[Optional[Set[types.TypeKey]]] = []
    for packed in self.checkout(ref_ids):
      if packed:
        # Add resource types.
        type_set = {
          value.type_info for value in utils.walk_resource_dict(packed.data)}
        # Add reference type.
        type_set.add(packed.ref_dict.type_info)
        result.append(type_set)
      else:
        result.append(None)
    return result

  def get_dependency_graph(
      self,
      ref_ids: Iterable[str],
      max_depth: Optional[int]=None) -> Dict[str, Optional[Set[str]]]:
    """Return the dependency graph for the input references.

    The default implementation will load all resource data and extract only
    the reference graph. This should be overridden if more efficiency is
    required.

    Args:
      ref_dict: The reference to retrieve dependencies for, encoded as a
        resource_dict.
      max_depth: The maximum depth to retrieve dependencies. If None, all
        dependencies will be retrieved.

    Returns:
      A dictionary from reference IDs to sets of references that the key
      depends on directly. An unknown reference will map to None. The set of
      keys in the returned dictionary will contain the IDs of the references
      in ref_dicts and all references that they depend on up to the specified
      depth.
    """
    result: Dict[str, Optional[Set[str]]] = {}

    # Set up BFS 'queue', which is going to be batch processed.
    seen: Set[str] = set(ref_ids)
    this_level = set(seen)
    depth = 0

    while this_level and (max_depth is None or depth <= max_depth):
      # Batch fetch all unfetched references at this depth.
      packed_resources = self.checkout(this_level)
      next_level: Set[str] = set()

      for ref_id, packed in zip(this_level, packed_resources):
        if packed is None:
          result[ref_id] = None
        else:
          result[ref_id] = packed.links
          next_level.update(packed.links - seen)
          seen.update(packed.links)

      # Update loop variables
      depth += 1
      this_level = next_level
    return result


class NotFound(Exception):
  """Raised when a resource is not found."""


class InMemoryBackend(StorageBackend):
  """A backend that stores resources in memory."""

  def __init__(self):
    """Create a new in-memory backend."""
    self._resources: Dict[str, PackedResource] = {}
    self._types: Dict[
      types.TypeKey, Dict[str, Optional[types.TypeDescriptor]]] = {}

  def register_type(self,
                    type_key: types.TypeKey,
                    attributes: Dict[str, Optional[types.TypeDescriptor]]):
    """Register a new type to allow resource to be committed."""
    if type_key in self._types and self._types[type_key] != attributes:
      raise TypeKeyError(
        f'Type {type_key} already registered with different attributes.')
    self._types[type_key] = attributes

  def get_type(self, type_key: types.TypeKey) -> (
      Optional[Dict[str, Optional[types.TypeDescriptor]]]):
    """Returns the type, if known."""
    return self._types[type_key]

  def commit(self, ref_id: str, packed_resource: PackedResource):
    """Stores a packed resource."""
    self._resources[ref_id] = packed_resource

  def has(self, ref_ids: Iterable[str]) -> List[bool]:
    """Returns whether the backend has the referenced resource."""
    return [ref_id in  self._resources for ref_id in ref_ids]

  def checkout(self, ref_ids: Iterable[str]) -> (
      List[Optional[PackedResource]]):
    """Returns a dictionary with resource data or None if not available."""
    return [self._resources.get(ref_id) for ref_id in ref_ids]

  def __len__(self) -> int:
    """Returns the number of resources in the backend."""
    return len(self._resources)


class FileBackend(StorageBackend):
  """A backend that stores resources in files."""

  def __init__(self,
               root_dir: str,
               serializer: Optional[serialization.Serializer] = None,
               use_base64_names: bool=True):
    """Create a new file-backed backend.

    Args:
      root_dir: The directory where resources will be stored.
      serialized: The serializer to use. Will default to JsonSerializer if not
        provided.
      use_base64_names: Use base64 encoded filenames for resources. This is
        useful on windows, since windows does not allow certain characters in
        file names.
    """
    os.makedirs(root_dir, exist_ok=True)
    self._root_dir = root_dir
    self._serializer = serializer or serialization.JsonSerializer()
    self._use_base64_names = use_base64_names

  def register_type(self,
                    type_key: types.TypeKey,
                    attributes: Dict[str, Optional[types.TypeDescriptor]]):
    """Register a new type to allow resource to be committed."""
    with open(self._get_type_path(type_key), 'wb') as file:
      encoded_attrs = {
        key: value.to_json() if value else None
        for key, value in attributes.items()}
      pickle.dump(encoded_attrs, file)

  def get_type(self, type_key: types.TypeKey) -> (
      Optional[Dict[str, Optional[types.TypeDescriptor]]]):
    """Returns the type, if known."""
    if not os.path.exists(self._get_type_path(type_key)):
      return None
    with open(self._get_type_path(type_key), 'rb') as f:
      encoded_attrs = pickle.load(f)
    assert isinstance(encoded_attrs, dict)
    return {
      key: types.TypeDescriptor.from_json(value)
      for key, value in encoded_attrs.items()}

  def _get_type_path(self, type_key: types.TypeKey) -> str:
    type_id = json.dumps(type_key.as_dict()).encode('utf-8')
    basename = f'type_{base64.b64encode(type_id).decode("utf-8")}'
    return os.path.join(self._root_dir, basename)

  def _get_path(self, ref_id: str) -> str:
    basename = ref_id
    if self._use_base64_names:
      basename = base64.b64encode(basename.encode('utf-8')).decode('utf-8')
    return os.path.join(self._root_dir, basename)

  def commit(self, ref_id: str, packed_resource: PackedResource):
    """Stores a packed resource."""
    data_bytes = self._serializer.serialize(packed_resource.data)
    ref_bytes = self._serializer.serialize(packed_resource.ref_dict)
    links = packed_resource.links
    with open(self._get_path(ref_id), 'wb') as file:
      pickle.dump((data_bytes, ref_bytes, links), file)

  def has(self, ref_ids: Iterable[str]) -> List[bool]:
    """Returns whether the backend has the referenced resource."""
    return [os.path.exists(self._get_path(ref_id)) for ref_id in ref_ids]

  def checkout(self, ref_ids: Iterable[str]) -> (
      List[Optional[PackedResource]]):
    """Returns a dictionary with resource data or None if not available."""
    return [self._get_packed(ref_id) for ref_id in ref_ids]

  def _get_packed(self, ref_id: str) -> Optional[PackedResource]:
    """Return the packed resource for a reference."""
    path = self._get_path(ref_id)
    if not os.path.exists(path):
      return None
    with open(path, 'rb') as file:
      data_bytes, ref_bytes, links = pickle.load(file)
    data: interfaces.ResourceDict = self._serializer.deserialize(data_bytes)
    ref_dict: interfaces.ResourceDict = self._serializer.deserialize(ref_bytes)
    return PackedResource(data, ref_dict, links)


class DistributionKeyError(Exception):
  """Raised when there are issues with distribution key objects.."""


class TypeKeyError(Exception):
  """Raised when there is an issue with type key objects."""



@contexts.register
class Store(contexts.Context):
  """A store for resources."""

  def __init__(
      self,
      backend: Optional[StorageBackend]=None,
      registry: Optional[resource_registry.Registry]=None,
      ref_type: Type[Ref]=Ref):
    """Initializes the store."""
    super().__init__()
    self._backend = backend if backend is not None else InMemoryBackend()
    self._registry = registry
    self._ref_type = ref_type

  def register_type(self, resource: Union[Type[R], R]):
    """Explicitly register a type with the store."""
    as_resource: Union[Type[interfaces.ResourceBase], interfaces.ResourceBase]
    if isinstance(resource, type):
      as_resource = resource_registry.wrap_type(resource)
    else:
      as_resource = resource_registry.wrap(resource)
    attributes = dict(
      zip(as_resource.field_names(), as_resource.field_descriptors()))
    self._backend.register_type(as_resource.type_key(), attributes)

  def commit(self, resource: R) -> Ref[R]:
    """Commits a resource to the store."""
    as_resource = resource_registry.wrap(resource)
    self.register_type(as_resource)
    ref, packed_resource = self._ref_type.pack(as_resource)
    self._backend.commit(ref.id, packed_resource)
    return ref

  def has(self, ref: Ref) -> bool:
    """Returns whether the store has a resource."""
    return self._backend.has((ref.id,))[0]

  def checkout(self, ref: Ref[R]) -> R:
    """Retrieves a resource from the store."""
    packed_resource = self._backend.checkout((ref.id,))[0]
    if packed_resource is None:
      raise NotFound(ref.id)
    if packed_resource.ref() != ref:
      raise RefError(
        f'Backend returned wrong reference:\ngot {packed_resource.ref()}\n'
        f'expected: {ref}')
    return ref.unpack(packed_resource)

  def get_transitive_type_requirements(self, ref: Ref) -> (
      Set[types.TypeKey]):
    """Return a set of transitive type requirements for the reference."""
    graph = self._backend.get_dependency_graph([ref.id])
    all_references = {ref.id}
    for ref_id, deps in graph.items():
      if deps is None:
        raise NotFound(f'Could not resolve transitive reference {ref_id}.')
      all_references.update(deps)
    type_sets = self._backend.get_type_keys(all_references)
    result: Set[types.TypeKey] = set()
    for typed_ref, type_set in zip(all_references, type_sets):
      if type_set is None:
        raise TypeKeyError(
          f'Could not resolve types for transitive reference {typed_ref}.')
      result.update(type_set)
    return result

  def get_distribution_requirements(
        self, ref: Ref, expect_distribution_key: bool=True) -> (
      Set[types.DistributionKey]):
    """Return the distribution requirements of a reference.

    Args:
      ref: The reference to check.
      expect_distribution_key: If True, raise an error if the reference does
        not have a distribution key.
    """
    type_requirements = self.get_transitive_type_requirements(ref)
    result: Set[types.DistributionKey] = set()
    for type_info in type_requirements:
      if type_info.distribution_key is None:
        if expect_distribution_key:
          raise DistributionKeyError(
            f'Type {type_info} does not have distribution key.')
      else:
        result.add(type_info.distribution_key)
    return result


@contexts.register_to_superclass(Store)
class FileStore(Store):
  """A file-based store."""

  def __init__(self, root_dir: str):
    """Initializes the store."""
    super().__init__(backend=FileBackend(root_dir))


@contexts.register_to_superclass(Store)
class InMemoryStore(Store):
  """An in-memory store."""

  def __init__(self):
    """Initializes the store."""
    super().__init__(backend=InMemoryBackend())



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Wrappers for python callables."""

import inspect
from types import MethodType
from typing import (
  Any, Callable, Dict, List, Mapping, Optional, Sequence, Tuple, Type, TypeVar,
  Union)
import dataclasses
import types

import wrapt  # type: ignore

from enact import interfaces
from enact import invocations
from enact import references
from enact import resource_registry
from enact import resources


C = TypeVar('C', bound=Callable)


@resource_registry.register
@dataclasses.dataclass
class CallArgs(resources.Resource):
  """Represents call arguments to a python function.

  This class is used for wrapping python functions.
  """
  args: List
  kwargs: Dict[str, Any]

  def to_python_args(self) -> Tuple[List, Dict[str, Any]]:
    """Transform into python call args."""
    return self.args, self.kwargs

  @classmethod
  def from_python_args(cls, *args, **kwargs) -> 'CallArgs':
    """Get CallArgs from python call args."""
    return cls(list(args), dict(kwargs))



def _create_wrapt_function_wrapper(
    enact_wrapper_type: Type[resource_registry.FunctionWrapper]):
  """Create a wrapt function wrapper for the given enact wrapper."""
  class WraptBoundFunctionWrapper(wrapt.BoundFunctionWrapper):

    def __init__(self, *args, **kwargs):
      super().__init__(*args, **kwargs)

    @property
    def _enact_function_wrapper_type(self) -> Type[
        resource_registry.FunctionWrapper]:
      return enact_wrapper_type

  class WraptFunctionWrapper(wrapt.FunctionWrapper):
    __bound_function_wrapper__ = WraptBoundFunctionWrapper

    @property
    def _enact_function_wrapper_type(self) -> Type[
        resource_registry.FunctionWrapper]:
      return enact_wrapper_type

  return WraptFunctionWrapper


def _wrapt_wrapper(
    enact_function_wrapper_type: Type[resource_registry.FunctionWrapper]):
  """A decorator that routes a function into an Invokable.

  This decorator will replace a callable with a wrapt wrapper
  that routes calls into an enact Invokable of the type indicated
  in enact_wrapper_type.

  Args:
    enact_wrapper_type: The type of the enact
  """
  def _wrapper_fun(wrapped, instance, args, kwargs):
    call_args = CallArgs.from_python_args(*args, **kwargs)
    wrapper: Union[resource_registry.FunctionWrapper,
                   resource_registry.MethodWrapper]
    if instance is not None:
      assert not isinstance(instance, type), (
        'Wrapping class methods is currently not supported.')
      wrapper = enact_function_wrapper_type.method_wrapper()(
        instance)  # type: ignore
      # If this function happens to be assigned as a member of the instance,
      # without being a proper 'member function', then we get a non-None
      # instance on an unbound function.
      wrapt_wrapped_func = getattr(wrapped, '__func__', wrapped)
    else:
      wrapper = enact_function_wrapper_type()
      wrapt_wrapped_func = wrapped
    assert wrapt_wrapped_func == wrapper.wrapper_function(), (
      f'Unexpected wrapped function passed in by wrapt: '
      f'{wrapt_wrapped_func} != {wrapper.wrapper_function()}')
    return wrapper(call_args)  # type: ignore
  return wrapt.decorator(
    wrapper=_wrapper_fun,
    proxy=_create_wrapt_function_wrapper(enact_function_wrapper_type))


class FunctionWrapper(invocations.Invokable,
                      resource_registry.FunctionWrapper):
  """Invokable base class for synchronous functions."""

  def call(self, call_args: CallArgs):
    args, kwargs = call_args.to_python_args()
    return self.wrapped_function()(*args, **kwargs)


@dataclasses.dataclass
class MethodWrapper(invocations.Invokable,
                    resource_registry.MethodWrapper):
  """Invokable base class for synchronous methods."""
  instance: Any

  def call(self, call_args: CallArgs):
    bound = types.MethodType(self.wrapped_function(),
                             self.get_instance())
    args, kwargs = call_args.to_python_args()
    return bound(*args, **kwargs)  # pylint: disable=not-callable


class AsyncFunctionWrapper(invocations.AsyncInvokable,
                           resource_registry.FunctionWrapper):
  """Invokable base class for synchronous functions."""

  async def call(self, call_args: CallArgs):
    args, kwargs = call_args.to_python_args()
    return await self.wrapped_function()(*args, **kwargs)


@dataclasses.dataclass
class AsyncMethodWrapper(invocations.AsyncInvokable,
                         resource_registry.MethodWrapper):
  """Invokable base class for synchronous methods."""
  instance: Any

  async def call(self, call_args: CallArgs):
    bound = MethodType(self.wrapped_function(),
                       self.get_instance())
    args, kwargs = call_args.to_python_args()
    return await bound(*args, **kwargs)  # pylint: disable=not-callable




def _create_async_function_wrapper(wrapped: Callable) -> (
    Type[AsyncFunctionWrapper]):
  """Create an enact async function wrapper for the given callable."""
  class _MethodWrapper(AsyncMethodWrapper):
    """The generated method wrapper."""
    # List valued to work around bug:
    # https://github.com/GrahamDumpleton/wrapt/issues/256
    _wrapper_function: List[Callable] = []

    @classmethod
    def wrapped_function(cls) -> Callable:
      """The associated function wrapper."""
      return wrapped

    @classmethod
    def wrapper_function(cls) -> Callable:
      """The associated function wrapper."""
      assert cls._wrapper_function
      return cls._wrapper_function[0]

  class _FunctionWrapper(AsyncFunctionWrapper):
    """The generated function wrapper."""
    # List valued to work around bug:
    # https://github.com/GrahamDumpleton/wrapt/issues/256
    _wrapper_function: List[Callable] = []

    @classmethod
    def wrapped_function(cls) -> Callable:
      """The function (not method) that is wrapped by this."""
      return wrapped

    @classmethod
    def wrapper_function(cls) -> Callable:
      """The associated function wrapper."""
      assert cls._wrapper_function
      return cls._wrapper_function[0]

    @classmethod
    def method_wrapper(cls) -> Type[_MethodWrapper]:
      """The associated method wrapper."""
      return _MethodWrapper

  return _FunctionWrapper


def _create_sync_function_wrapper(fun: Callable) -> (
    Type[FunctionWrapper]):
  """Create an enact function wrapper for the given callable."""
  class _MethodWrapper(MethodWrapper):
    """The generated method wrapper"""
    # List valued to work around bug:
    # https://github.com/GrahamDumpleton/wrapt/issues/256
    _wrapper_function: List[Callable] = []

    @classmethod
    def wrapped_function(cls) -> Callable:
      """The associated wrapped function."""
      return fun

    @classmethod
    def wrapper_function(cls) -> Callable:
      """The associated wrapper function."""
      assert cls._wrapper_function
      return cls._wrapper_function[0]

  class _FunctionWrapper(FunctionWrapper):
    """The generated function wrapper."""
    # List valued to work around bug:
    # https://github.com/GrahamDumpleton/wrapt/issues/256
    _wrapper_function: List[Callable] = []

    @classmethod
    def wrapped_function(cls) -> Callable:
      """The associated wrapped function."""
      return fun

    @classmethod
    def wrapper_function(cls) -> Callable:
      """The associated wrapper function."""
      assert cls._wrapper_function
      return cls._wrapper_function[0]  # type: ignore

    @classmethod
    def method_wrapper(cls) -> Type[_MethodWrapper]:
      """The associated method wrapper."""
      return _MethodWrapper

  return _FunctionWrapper


def _create_function_wrapper(fun: Callable) -> (
    Type[resource_registry.FunctionWrapper]):
  """Create a sync or async function wrapper for the callable."""
  function_wrapper_type: Union[
    Type[FunctionWrapper],
    Type[AsyncFunctionWrapper]]
  if inspect.iscoroutinefunction(fun):
    function_wrapper_type = _create_async_function_wrapper(fun)
  else:
    function_wrapper_type = _create_sync_function_wrapper(fun)
  # Override names and qualnames to match the wrapped function.
  for generated_type in (function_wrapper_type,
                         function_wrapper_type.method_wrapper()):
    generated_type.__module__ = fun.__module__
    generated_type.__name__ = fun.__name__
    generated_type.__qualname__ = fun.__qualname__
  function_wrapper_type.method_wrapper().__name__ += '__method_wrapper'
  function_wrapper_type.method_wrapper().__qualname__ += '__method_wrapper'
  return function_wrapper_type


def register(fun: C) -> C:
  """Decorator for registering a wrapped function."""
  if inspect.ismethod(fun):
    raise ValueError(
      'Bound methods (e.g., A().foo) cannot be registered. '
      'Please register the unbound function instead (e.g., A.foo)')
  wrapper_type = _create_function_wrapper(fun)
  wrapper_fun = _wrapt_wrapper(wrapper_type)(fun)

  # pylint: disable=protected-access
  wrapper_type._wrapper_function = [wrapper_fun]  # type: ignore
  wrapper_type.method_wrapper()._wrapper_function = [  # type: ignore
    wrapper_fun]  # type: ignore

  resource_registry.register(wrapper_type)
  resource_registry.register(wrapper_type.method_wrapper())
  return wrapper_fun

@resource_registry.register
@dataclasses.dataclass
class Call(invocations.Invokable[CallArgs, Any]):
  """An invokable representing a function call to an unregistered function."""
  function_name: str

  def __post_init__(self):
    """"""
    self._callable: Optional[Callable] = None

  @classmethod
  def create(cls, function_name: str, fun: Callable) -> 'Call':
    """Create a new call, setting the callable."""
    call = Call(function_name)
    call._callable = fun
    return call

  def call(self, resource: CallArgs) -> Any:
    """Call the wrapped function."""
    if not self._callable:
      raise invocations.InvocationError(
        'Cannot repeat a call of an unregistered function. '
        'Please make sure your top-level function is registered.')
    args, kwargs = resource.to_python_args()
    return self._callable(*args, **kwargs)

@resource_registry.register
@dataclasses.dataclass
class AsyncCall(invocations.AsyncInvokable[CallArgs, Any]):
  """An invokable representing a call to an async unregistered function."""
  function_name: str

  def __post_init__(self):
    """"""
    self._callable: Optional[Callable] = None

  @classmethod
  def create(cls, function_name: str, fun: Callable) -> 'AsyncCall':
    """Create a new call, setting the callable."""
    call = AsyncCall(function_name)
    call._callable = fun
    return call

  async def call(self, resource: CallArgs) -> Any:
    """Call the wrapped function."""
    if not self._callable:
      raise invocations.InvocationError(
        'Cannot repeat a call of an unregistered function. '
        'Please make sure your top-level function is registered.')
    args, kwargs = resource.to_python_args()
    return await self._callable(*args, **kwargs)


def get_wrapper_type(fun: Callable) -> Optional[
    resource_registry.FunctionWrapper]:
  """Return the CallableInvokable of this function, if it has one."""
  return getattr(fun, '_enact_function_wrapper_type', None)


class _NotAnInvokable(Exception):
  """Raised when a callable is not an invokable."""


def _get_invokable_and_arg(
  fun: Callable, args: Sequence, kwargs: Mapping[str, Any]) -> Tuple[
    Callable, invocations._InvokableBase, references.Ref]:
  """Get the underlying invokable and relevant resource arg."""
  try:
    wrapped = resource_registry.wrap(fun)  # Try wrapping the callable.
  except interfaces.FieldTypeError:
    wrapped = None
  call_args: Optional[references.Ref] = None
  if wrapped:
    if not callable(wrapped):
      raise TypeError(
        f'The wrapper {wrapped} of {fun} is not callable.')
    fun = wrapped

  # pylint: disable=protected-access
  if not isinstance(fun, invocations._InvokableBase):
    raise _NotAnInvokable()

  invokable = fun
  if fun.get_input_type() == CallArgs:
    call_args = references.commit(CallArgs.from_python_args(
      *args, **kwargs))
  elif isinstance(fun, (resource_registry.FunctionWrapper,
                        resource_registry.MethodWrapper)):
    call_args = references.commit(CallArgs.from_python_args(
      *args, **kwargs))
  else:
    if len(args) > 1 or kwargs:
      raise ValueError(
        'Invokables can only be invoked with up to one arg and no kwargs.')
    if args:
      call_args = references.commit(args[0])
    else:
      call_args = references.commit(None)
  return fun, invokable, call_args


def _strip_extra_call(
    fun: Callable,
    invocation: invocations.Invocation,
    commit: bool) -> invocations.Invocation:
  """Strip extraneous Call invokable."""
  # Remove the extraneous Call invocation if the called object
  # was a registered python function.
  # pylint: disable=protected-access
  if not isinstance(fun, invocations._InvokableBase):
    # Remove the top level of invocation.
    invokable_wrapper = get_wrapper_type(fun)
    if invokable_wrapper:
      children = list(invocation.get_children())
      assert len(children) == 1, (
        f'Expected one child containing the CallableInvokable associated '
        f'with {fun}. Got {len(children)} children instead.')
      invocation = children[0]
  elif commit:
    # This is an elif rather than an if, because the child is already
    # committed regardless of the setting, so we don't need to repeat
    # work in that case.
    references.commit(invocation)
  return invocation


def invoke(
    fun: Callable,
    args: Sequence=(),
    kwargs: Optional[Mapping[str, Any]]=None,
    replay_from: Optional[invocations.Invocation]=None,
    exception_override: invocations.ExceptionOverride=lambda _: None,
    raise_on_errors: Tuple[Type[Exception], ...]=(
      invocations.InvocationError, interfaces.FrameworkError),
    wrap_exceptions: bool=False,
    strict: bool=True,
    commit: bool=True) -> invocations.Invocation:
  """Invoke an invokable, wrapped function or python callable.

  Args:
    fun: A callable to invoke.
    args: The args to pass to callable.
    kwargs: The kwargs to pass to callable.
    replay_from: The invocation to replay from.
    exception_override: A function that takes an exception and returns either
      None or a resource to override the exception with.
    raise_on_errors: Which errors should raise beyond the invocation.
    wrap_exceptions: Whether to wrap non-resource exceptions in NativeException.
      If False, non-resource exceptions raised during invocation will be
      re-raised.
    strict: If replaying, whether to replay strictly (expecting the same calls).
    commit: Whether to commit the invocation after invoking.
  """
  kwargs = kwargs or {}
  resource_arg: Optional[references.Ref] = None
  try:
    fun, invokable, resource_arg = _get_invokable_and_arg(fun, args, kwargs)
  except _NotAnInvokable:
    resource_arg = references.commit(
      CallArgs.from_python_args(*args, **kwargs))
    invokable = Call.create(f'{fun.__module__}.{fun.__qualname__}', fun)

  if isinstance(invokable, invocations.AsyncInvokableBase):
    raise TypeError(
      'Cannot invoke an AsyncInvokable. Use invoke_async instead.')

  assert isinstance(invokable, invocations.InvokableBase)

  invocation = invokable.invoke(
    resource_arg,
    replay_from=replay_from,
    exception_override=exception_override,
    raise_on_errors=raise_on_errors,
    wrap_exceptions=wrap_exceptions,
    strict=strict,
    # Skip commit, since we may throw away the outermost invocation:
    commit=False)

  return _strip_extra_call(fun, invocation, commit)


async def invoke_async(
    fun: Callable,
    args: Sequence=(),
    kwargs: Optional[Mapping[str, Any]]=None,
    replay_from: Optional[invocations.Invocation]=None,
    exception_override: invocations.ExceptionOverride=lambda _: None,
    raise_on_errors: Tuple[Type[Exception], ...]=(
      invocations.InvocationError, interfaces.FrameworkError),
    wrap_exceptions: bool=False,
    strict: bool=True,
    commit: bool=True) -> invocations.Invocation:
  """Invoke an invokable, wrapped function or python callable.

  Args:
    fun: A callable to invoke.
    args: The args to pass to callable.
    kwargs: The kwargs to pass to callable.
    replay_from: The invocation to replay from.
    exception_override: A function that takes an exception and returns either
      None or a resource to override the exception with.
    raise_on_errors: Which errors should raise beyond the invocation.
    wrap_exceptions: Whether to wrap non-resource exceptions in NativeException.
      If False, non-resource exceptions raised during invocation will be
      re-raised.
    strict: If replaying, whether to replay strictly (expecting the same calls).
    commit: Whether to commit the invocation after invoking.
  """
  kwargs = kwargs or {}
  resource_arg: Optional[references.Ref] = None
  try:
    fun, invokable, resource_arg = _get_invokable_and_arg(fun, args, kwargs)
  except _NotAnInvokable:
    resource_arg = references.commit(
      CallArgs.from_python_args(*args, **kwargs))
    invokable = AsyncCall.create(f'{fun.__module__}.{fun.__qualname__}', fun)

  if isinstance(invokable, invocations.InvokableBase):
    raise TypeError(
      'Cannot invoke_async an Invokable. Use invoke instead.')

  assert isinstance(invokable, invocations.AsyncInvokableBase)

  invocation = await invokable.invoke(
    resource_arg,
    replay_from=replay_from,
    exception_override=exception_override,
    raise_on_errors=raise_on_errors,
    wrap_exceptions=wrap_exceptions,
    strict=strict,
    # Skip commit, since we may throw away the outermost invocation:
    commit=False)

  return _strip_extra_call(fun, invocation, commit)



# Copyright 2024 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Context to guard against cyclic datastructures."""

from typing import List, Optional
from enact import contexts


class CycleDetected(Exception):
  """Exception raised when a cycle is detected."""


@contexts.register
class AcyclicContext(contexts.Context):
  """Helper to safeguard against cyclic data-structures."""

  def __init__(self, obj):
    """Initializes the context."""
    super().__init__()
    self.parent: Optional[AcyclicContext] = None
    self.obj = obj

  def enter(self):
    """Check if the value is already on the stack."""
    self.parent = AcyclicContext.get_current()
    parent = self.parent
    parents: List[AcyclicContext] = []
    while parent is not None:
      parents.append(parent)
      if parent.obj is self.obj:
        raise CycleDetected(
          f'Resources may not have cyclic graph structure. '
          f'Encountered cycle: '
          f'{" -> ".join(str(p.obj) for p in parents)}')
      parent = parent.parent



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests registration module."""

from typing import Optional, Union
import unittest

import enact

class RegistrationTest(unittest.TestCase):
  """Tests for the enact.register function."""

  def test_type_inference(self):
    """Tests invokable type inference."""
    @enact.register
    class MyInvokable(enact.Invokable):
      """Invokable for testing."""

      def call(self, x: int) -> str:
        return str(x)

    self.assertEqual(MyInvokable.get_input_type(), int)
    self.assertEqual(MyInvokable.get_output_type(), str)

  def test_type_inference_none_return(self):
    """Tests invokable type inference if none is returned."""
    @enact.register
    class MyInvokable(enact.Invokable):
      """Invokable for testing."""

      def call(self, x: int) -> None:
        pass

    self.assertEqual(MyInvokable.get_input_type(), int)
    self.assertEqual(MyInvokable.get_output_type(), type(None))

  def test_type_inference_no_type(self):
    """Tests invokable type inference if no types are set."""
    @enact.register
    class MyInvokable(enact.Invokable):
      """Invokable for testing."""

      def call(self, x):
        pass

    self.assertEqual(MyInvokable.get_input_type(), None)
    self.assertEqual(MyInvokable.get_output_type(), None)

  def test_type_inference_union(self):
    """Tests type inference fails for union and optional types."""
    @enact.register
    class MyInvokable(enact.Invokable):
      """Invokable for testing."""

      def call(self, x: Union[int, str]) -> Optional[str]:
        pass
    self.assertEqual(MyInvokable.get_input_type(), None)
    self.assertEqual(MyInvokable.get_output_type(), None)

  def test_type_inference_no_arg(self):
    """Tests invokable type inference if the call function has no args."""
    @enact.register
    class MyInvokable(enact.Invokable):
      """Invokable for testing."""
      def call(self):
        pass

    self.assertEqual(MyInvokable.get_input_type(), type(None))
    self.assertEqual(MyInvokable.get_output_type(), None)

  def test_register_too_many_args(self):
    """Test too many argument errors on invokables."""
    class TooManyArgs(enact.Invokable):
      """Invokable for testing."""
      def call(self, x, y):
        pass
    with self.assertRaises(TypeError):
      enact.register(TooManyArgs)

  def test_register_keyword_only_args(self):
    """Test wrong kind error on invokables."""
    class KeywordOnly(enact.Invokable):
      """Invokable for testing."""
      def call(self, *, x):
        pass
    with self.assertRaises(TypeError):
      enact.register(KeywordOnly)

  def test_register_no_self_arg(self):
    """Test wrong kind error on invokables."""
    class StaticCall(enact.Invokable):
      """Invokable for testing."""
      @staticmethod
      def call(x):
        pass
    with self.assertRaises(TypeError):
      enact.register(StaticCall)

  def test_register_type_not_resource(self):
    """Test type error on non-resource."""
    class NotAResource:
      """A class that is not a resource."""
      pass
    with self.assertRaises(TypeError):
      enact.register(NotAResource)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the serialize module."""
import dataclasses
from typing import Any, Dict, List, Type
import unittest

import enact
from enact import serialization
from enact import resource_registry

import random_value  # type: ignore

@enact.register
@dataclasses.dataclass
class AllTypesResource(enact.Resource):
  """A resource that contains all valid field types."""
  i: int
  f: float
  bl: bool
  b: bytes
  s: str
  n: Any
  r: enact.Ref
  m: Dict
  l: List
  t: Type[enact.Resource]


class JsonSerializerTest(unittest.TestCase):
  """Tests the JSON serializer."""

  def setUp(self):
    """Sets up the test case."""
    self.registry = resource_registry.Registry()
    self.registry.register(enact.Ref)
    self.serializer = serialization.JsonSerializer(self.registry)

  def test_serialize_deserialize(self):
    """Tests that serialization/deserialization works."""
    self.registry.register(AllTypesResource)
    resource = AllTypesResource(
      i=2, f=3.0, bl=True, b=b'bytes', s='test', n=None,
      r=enact.Ref('12314'), m={'a': ['test']}, l=[{'b': 1}, {'c': 2}],
      t=AllTypesResource)
    got = self.serializer.serialize(resource.to_resource_dict())
    deserialized = resource_registry.from_resource_dict(
      self.serializer.deserialize(got))
    self.assertEqual(deserialized, resource)

  def test_serialize_deserialize_fuzz(self):
    """Fuzz test serialization and deserialization."""
    self.registry.register(random_value.R)
    for _ in range(100):
      resource = random_value.rand_resource()
      got = self.serializer.serialize(resource.to_resource_dict())
      deserialized = resource_registry.from_resource_dict(
        self.serializer.deserialize(got))
      self.assertEqual(deserialized, resource)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Type registration functionality to allow deserialization of resources."""

import inspect
import types as types_module
from typing import (
  Any, Callable, Dict, Hashable, Iterable, List, Mapping, Optional, Set,
  Type, TypeVar, Union, cast)

from enact import distribution_registry
from enact import interfaces
from enact import types



WrappedT = TypeVar('WrappedT')
WrapperT = TypeVar('WrapperT', bound=interfaces.TypeWrapperBase)
ResourceT = TypeVar('ResourceT', bound=interfaces.ResourceBase)

FieldValueWrapperT = TypeVar('FieldValueWrapperT', bound='FieldValueWrapper')
FunctionWrapperT = TypeVar('FunctionWrapperT', bound='FunctionWrapper')
MethodWrapperT = TypeVar('MethodWrapperT', bound='MethodWrapper')

class RegistryError(Exception):
  """Raised when there is an error with the registry."""


class ResourceNotFound(RegistryError):
  """Raised when a resource is not found."""


class FunctionWrapper(interfaces.ResourceBase):
  """Base class for function wrappers.

  Should be subclassed as an Invokable or AsyncInvokable.

  Function wrapping require a three step approach:
  1) The native function is transparently wrapped using a python wrapper
     function.
  2) The wrapper function routes calls into an invokable resource, so that
     inputs are tracked.
  3) The invokable calls the native function that is being wrapped.

  The native function that is being wrapped is exposed as wrapped_function.
  The python wrapper function that calls into the invokable is exposed as
  wrapper function.
  """
  @classmethod
  def wrapper_function(cls) -> Callable:
    """The python function that routes calls into this invokable."""
    raise NotImplementedError()

  @classmethod
  def wrapped_function(cls) -> Callable:
    """The python function that is being called by this invokable."""
    raise NotImplementedError()

  @classmethod
  def wrap(cls: Type[FunctionWrapperT], c: Callable) -> Union[
      'FunctionWrapper', 'MethodWrapper']:
    """Wrap a function."""
    if inspect.ismethod(c):
      return cls.method_wrapper().wrap(c)
    assert c == cls.wrapper_function(), (
      f'Wrong wrapper class for callable: {cls} vs {c}')
    return cls()

  @classmethod
  def method_wrapper(cls) -> 'Type[MethodWrapper]':
    """The associated method wrapper."""
    raise NotImplementedError()


class MethodWrapper(interfaces.ResourceBase):
  """Base class for method wrappers."""

  @classmethod
  def wrap(cls: Type[MethodWrapperT],
           m: types_module.MethodType) -> MethodWrapperT:
    """Wrap a method."""
    assert inspect.ismethod(m), 'Expected method.'
    assert m.__self__, 'Expected __self__ attribute on bound method.'
    assert m.__func__ == cls.wrapper_function(), (
      'Expected method to be based on the same function as its '
      'function wrapper.')
    # pylint: disable=too-many-function-args
    return cls(m.__self__) # type: ignore

  @classmethod
  def wrapper_function(cls) -> Callable:
    """The python function that routes calls into this invokable."""
    raise NotImplementedError()

  @classmethod
  def wrapped_function(cls) -> Callable:
    """The python function that is being called by this invokable."""
    raise NotImplementedError()

  def get_instance(self) -> Any:
    """Added to interface to simplify typing."""
    return self.instance  # type: ignore


class MissingWrapperError(interfaces.FieldTypeError):
  """Raised when a required wrapper is missing."""


class Registry:
  """Registers resource types for deserialization."""

  _singleton: Optional['Registry'] = None

  def __init__(self):
    """Initializes a registry."""
    # Ensure that the enact distribution is registered to its semantic version
    # before registering any resource types.
    distribution_registry.ensure_enact_registered()
    self.allow_reregistration = True

    # Map from type id to resource type.
    self._type_map: Dict[str, Type[interfaces.ResourceBase]] = {}

    # Map from python types to wrapper types.
    self._wrapped_types: Dict[Type, Type[interfaces.TypeWrapperBase]] = {}
    self._wrapper_types: Set[Type[interfaces.TypeWrapperBase]] = set()
    self._function_wrappers: Dict[Callable, Type[FunctionWrapper]] = {}


  def _from_dict_value(self, value: interfaces.ResourceDictValue) -> (
      interfaces.FieldValue):
    """Transforms a resource dict value to a field value."""
    if isinstance(value, types.PRIMITIVES):
      return value
    if isinstance(value, type) and issubclass(value, interfaces.ResourceBase):
      return value
    if isinstance(value, List):
      return [self._from_dict_value(x) for x in value]
    if isinstance(value, interfaces.ResourceDict):
      return self.from_resource_dict(value)
    if isinstance(value, Dict):
      def _assert_str(maybe_str: str) -> str:
        if type(maybe_str) is not str:  # pylint: disable=unidiomatic-typecheck
          raise interfaces.FieldTypeError(
            f'Expected string key, got {type(maybe_str)}')
        return maybe_str
      return {
        _assert_str(k): self._from_dict_value(v)
        for k, v in value.items()}
    raise interfaces.FieldTypeError(
      f'Encountered unsupported resource '
      f'dict value type {type(value)}: {value}')

  def deepcopy(self, resource: ResourceT) -> ResourceT:
    """Create a deep-copy of the resource."""
    return cast(ResourceT, self.from_resource_dict(resource.to_resource_dict()))

  def from_resource_dict(self, resource_dict: interfaces.ResourceDict) -> (
      interfaces.ResourceBase):
    """Constructs the resource from a ResourceDict dictionary."""
    if not isinstance(resource_dict, interfaces.ResourceDict):
      raise TypeError(f'Input is not a ResourceDict: {resource_dict}')
    resource_type = self.lookup(resource_dict.type_info)
    field_dict = {
      k: self._from_dict_value(v)
      for k, v in resource_dict.items()}
    return resource_type.from_fields(field_dict)

  def register(self, resource: Type[interfaces.ResourceBase]):
    """Registers the resource type."""
    # Check argument type.
    if not issubclass(resource, interfaces.ResourceBase):
      raise RegistryError(
        f'Cannot register non-resource type: {resource}')
    # Auto-add distribution key.
    dist_key = resource.type_distribution_key()
    if dist_key is None:
      dist_key = distribution_registry.get_distribution_key(resource)
      if dist_key is not None:
        resource.set_type_distribution_key(dist_key)
    # Record the type.
    type_id = resource.type_id()
    if (type_id in self._type_map and
        self._type_map[type_id] != resource and
        not self.allow_reregistration):
      raise RegistryError(
        f'{id(resource)} == {id(self._type_map[type_id])}\n'
        f'While registering {resource}: '
        f'Type with id {type_id} already '
        f'registered to a different type: '
        f'{self._type_map[type_id]}\n')
    self._type_map[type_id] = resource
    # Handle special types.
    if issubclass(resource, interfaces.TypeWrapperBase):
      self._register_type_wrapper(resource)
    if issubclass(resource, FunctionWrapper):
      return self._register_function_wrapper(resource)

  def lookup(self, type_id: Union[str, types.TypeKey]) -> (
      Type[interfaces.ResourceBase]):
    """Looks up a resource type by name or type_info."""
    if isinstance(type_id, types.TypeKey):
      type_id = type_id.type_id()
    resource_class = self._type_map.get(type_id)
    if not resource_class:
      raise ResourceNotFound(f'No type registered for {type_id}')
    return resource_class

  def _register_function_wrapper(self, wrapper_type: Type[FunctionWrapper]):
    """Register a function wrapper type."""
    self._function_wrappers[
        wrapper_type.wrapper_function()] = wrapper_type

  def _register_type_wrapper(self, wrapper_type: Type[WrapperT]):
    """Register a new wrapper type."""
    self._wrapped_types[wrapper_type.wrapped_type()] = wrapper_type
    self._wrapper_types.add(wrapper_type)

  def get_type_wrapper(self, t: Type[WrappedT]) -> Optional[
      Type[interfaces.TypeWrapperBase[WrappedT]]]:
    """Return a matching wrapper type if present."""
    wrapper = self._wrapped_types.get(t)
    if wrapper:
      return cast(Type[interfaces.TypeWrapperBase[WrappedT]], wrapper)
    found: Optional[Type[interfaces.TypeWrapperBase]] = None
    for k, v in self._wrapped_types.items():
      if issubclass(t, k):
        if found:
          raise RegistryError(
            f'Found multiple wrappers for type {t}: {found} and {v}')
        found = v
    return found

  def _get_function_wrapper_type(self, c: Callable) -> Type['FunctionWrapper']:
    """Return the function wrapper for c or raise an error."""
    func = c
    if inspect.ismethod(c):
      func = c.__func__
    if not isinstance(func, Hashable):
      raise interfaces.FieldTypeError(
        'Only immutable callables (e.g. python functions) can be handled '
        'using enact. If you need use callable python classes that are '
        'mutable, consider subclassing from Invokable or AsyncInvokable.')
    function_wrapper_type = self._function_wrappers.get(func)
    if not function_wrapper_type:
      raise MissingWrapperError(
        f'Cannot find wrapper for {func}. Did you register the '
        f'function or method with the @enact.register decorator?')
    return function_wrapper_type

  def wrap(self, value: Any) -> interfaces.ResourceBase:
    """Wrap a value if necessary."""
    if isinstance(value, interfaces.ResourceBase):
      return value
    type_wrapper_class = self.get_type_wrapper(type(value))
    if type_wrapper_class:
      return type_wrapper_class.wrap(value)
    if callable(value):
      return self._get_function_wrapper_type(value).wrap(value)
    if inspect.iscoroutine(value):
      raise MissingWrapperError(
        f'Cannot wrap coroutine {value}. Did you forget to await it?')
    raise MissingWrapperError(
      f'Cannot wrap value of type {type(value)}. Please register '
      f'a TypeWrapper for this type.')

  def wrap_type(self, value: Type) -> Type[interfaces.ResourceBase]:
    """Wrap a type if necessary."""
    if issubclass(value, interfaces.ResourceBase):
      return value
    wrapper = self.get_type_wrapper(value)
    if wrapper:
      return wrapper
    raise MissingWrapperError(
      f'Cannot wrap type {value}. Please register '
      f'a TypeWrapper for this type.')

  def unwrap(self, value: Any) -> Any:
    """Unwrap a value if wrapped."""
    if isinstance(value, interfaces.TypeWrapperBase):
      return value.unwrap()
    if isinstance(value, FunctionWrapper):
      return value.wrapper_function()
    if isinstance(value, MethodWrapper):
      return types_module.MethodType(
        value.wrapper_function(), value.get_instance())
    return value

  def unwrap_type(self, value: Type[interfaces.ResourceBase]) -> Type:
    """Unwrap a type if wrapped."""
    if issubclass(value, interfaces.TypeWrapperBase):
      return value.wrapped_type()
    return value

  @classmethod
  def get(cls) -> 'Registry':
    """Returns the singleton registry."""
    if not cls._singleton:
      cls._singleton = cls()
    return cls._singleton


def register(cls: Type[ResourceT]) -> Type[ResourceT]:
  """Decorator for resource classes."""
  Registry.get().register(cls)
  return cls


def register_wrapper(cls: Type[WrapperT]) -> Type[WrapperT]:
  """Decorator for resource wrapper classes."""
  # pylint: disable=protected-access
  Registry.get()._register_type_wrapper(cls)
  return cls


def wrap(value: Any) -> interfaces.ResourceBase:
  """Wrap a value as a resource if necessary."""
  return Registry.get().wrap(value)


def unwrap(value: Any) -> Any:
  """Unwrap a value if wrapped."""
  return Registry.get().unwrap(value)


def wrap_type(value: Type) -> Type[interfaces.ResourceBase]:
  """Wrap a type as a resource."""
  return Registry.get().wrap_type(value)


def unwrap_type(value: Type[interfaces.ResourceBase]) -> Type:
  """Wrap a type."""
  return Registry.get().unwrap_type(value)


def to_python_type(type_descriptor: types.TypeDescriptor) -> Type:
  """Convert a type descriptor to a python type."""
  if isinstance(type_descriptor, types.Int):
    return int
  if isinstance(type_descriptor, types.Str):
    return str
  if isinstance(type_descriptor, types.Float):
    return float
  if isinstance(type_descriptor, types.Bool):
    return bool
  if isinstance(type_descriptor, types.Bytes):
    return bytes
  if isinstance(type_descriptor, types.List):
    return list
  if isinstance(type_descriptor, types.Dict):
    return dict
  if isinstance(type_descriptor, types.ResourceType):
    resource_type = Registry.get().lookup(type_descriptor.type_key)
    if issubclass(resource_type, interfaces.TypeWrapperBase):
      return resource_type.wrapped_type()
    return resource_type
  raise ValueError(f'Unknown type descriptor: {type_descriptor}')

def from_python_type(t: Type) -> types.TypeDescriptor:
  """Convert a python type to an enact type descriptor."""
  if t == int:
    return types.Int()
  if t == str:
    return types.Str()
  if t == float:
    return types.Float()
  if t == bool:
    return types.Bool()
  if t == bytes:
    return types.Bytes()
  if t == list:
    return types.List()
  if t == dict:
    return types.Dict()
  if issubclass(t, interfaces.ResourceBase):
    return types.ResourceType(t.type_key())
  try:
    return types.ResourceType(wrap_type(t).type_key())
  except MissingWrapperError as e:
    raise ValueError(f'Cannot convert type {t} to a type descriptor.') from e


def deepcopy(value: WrappedT) -> WrappedT:
  """Deep copy a value."""
  if isinstance(value, interfaces.ResourceBase):
    result = from_resource_dict(value.to_resource_dict())
  else:
    result = unwrap(from_resource_dict(wrap(value).to_resource_dict()))
  return cast(WrappedT, result)


def _ensure_str_key(s: Any) -> str:
  """Ensure that a value is a string."""
  if not isinstance(s, str):
    raise ValueError(
      f'Cannot auto-wrap a dictionary with non-str keys: '
      f'{s} of type {type(s)}')
  return s


def to_field_value(value: Any) -> interfaces.FieldValue:
  """Wrap a value as a field value."""
  if isinstance(value, types.PRIMITIVES):
    return value
  if isinstance(value, list):
    return [to_field_value(x) for x in value]
  if isinstance(value, dict):
    return {_ensure_str_key(k): to_field_value(v) for k, v in value.items()}
  if isinstance(value, type):
    return Registry.get().wrap_type(value)
  return wrap(value)


def from_field_value(value: interfaces.FieldValue) -> Any:
  """Unwrap a field value into a python value."""
  if isinstance(value, interfaces.ResourceBase):
    return unwrap(value)
  if (
      isinstance(value, type) and
      issubclass(value, interfaces.TypeWrapperBase)):
    return unwrap_type(value)
  if isinstance(value, list):
    return [from_field_value(x) for x in value]
  if isinstance(value, dict):
    return {k: from_field_value(v) for k, v in value.items()}
  return value

def from_resource_dict(resource_dict: interfaces.ResourceDict) -> (
    interfaces.ResourceBase):
  """Constructs the resource from a ResourceDict dictionary."""
  return Registry.get().from_resource_dict(resource_dict)


class FieldValueWrapper(interfaces.TypeWrapperBase[WrappedT]):
  """Base class for field value wrappers."""
  value: interfaces.FieldValue

  def __init__(self, value: interfaces.FieldValue):
    """Initializes a wrapped field value."""
    self.value = value

  @classmethod
  def field_names(cls) -> Iterable[str]:
    """Returns the names of the fields of the resource."""
    return ('wrapped',)

  def field_values(self) -> Iterable[interfaces.FieldValue]:
    return (self.value,)

  @classmethod
  def from_fields(
      cls: Type[FieldValueWrapperT],
      field_dict: Mapping[str, interfaces.FieldValue]) -> FieldValueWrapperT:
    assert len(field_dict) == 1
    return cls(field_dict['wrapped'])

  def set_from(self, other: interfaces.ResourceBase):
    """Sets the fields of this resource from another resource."""
    if not type(self) is type(other):
      raise RegistryError(
        f'Cannot set fields from {type(other)} to {type(self)}.')
    assert isinstance(other, FieldValueWrapper)
    self.wrapped = deepcopy(other).value

  @classmethod
  def wrap(cls: Type[FieldValueWrapperT],
           value: WrappedT) -> FieldValueWrapperT:
    """Wrap a value directly."""
    assert isinstance(value, cls.wrapped_type()), (
      'Cannot wrap value of type {type(value)} with wrapper {cls}.')
    return cls(to_field_value(value))

  def unwrap(self) -> WrappedT:
    """Unwrap a value directly."""
    assert isinstance(self.value, self.wrapped_type())
    return from_field_value(self.value)


@register
class NoneWrapper(
  interfaces.TypeWrapperBase):
  """Wrapper for None."""
  @classmethod
  def wrapped_type(cls) -> Type[None]:
    return type(None)

  @classmethod
  def field_names(cls) -> Iterable[str]:
    """Returns the names of the fields of the resource."""
    return ()

  def field_values(self) -> Iterable[interfaces.FieldValue]:
    return ()

  @classmethod
  def from_fields(
      cls, field_dict: Mapping[str, interfaces.FieldValue]) -> 'NoneWrapper':
    assert len(field_dict) == 0
    return cls()

  def set_from(self, other: interfaces.ResourceBase):
    """Sets the fields of this resource from another resource."""
    if not type(self) is type(other):
      raise RegistryError(
        f'Cannot set fields from {type(other)} to {type(self)}.')

  @classmethod
  def wrap(cls, value: WrappedT) -> 'NoneWrapper':
    """Wrap a value directly."""
    assert isinstance(value, cls.wrapped_type()), (
      'Cannot wrap value of type {type(value)} with wrapper {cls}.')
    return NoneWrapper()

  def unwrap(self) -> None:
    """Unwrap a value directly."""
    return None

class PrimitiveWrapper(FieldValueWrapper[WrappedT]):
  """Wrapper for primitives."""

  @classmethod
  def is_immutable(cls) -> bool:
    return True


@register
class IntWrapper(PrimitiveWrapper[int]):
  """Wrapper for ints."""
  @classmethod
  def wrapped_type(cls) -> Type[int]:
    return int


@register
class FloatWrapper(PrimitiveWrapper[float]):
  """Wrapper for floats."""
  @classmethod
  def wrapped_type(cls) -> Type[float]:
    return float


@register
class BoolWrapper(PrimitiveWrapper[bool]):
  """Wrapper for bools."""
  @classmethod
  def wrapped_type(cls) -> Type[bool]:
    return bool


@register
class StrWrapper(PrimitiveWrapper[str]):
  """Wrapper for strs."""
  @classmethod
  def wrapped_type(cls) -> Type[str]:
    return str


@register
class BytesWrapper(PrimitiveWrapper[bytes]):
  """Wrapper for bytes."""
  @classmethod
  def wrapped_type(cls) -> Type[bytes]:
    return bytes


@register
class ListWrapper(FieldValueWrapper[list]):
  """Wrapper for lists."""
  @classmethod
  def wrapped_type(cls) -> Type[list]:
    return list

  @classmethod
  def set_wrapped_value(cls, target: list, src: list):
    target.clear()
    target.extend(src)


@register
class DictWrapper(FieldValueWrapper[dict]):
  """Wrapper for dicts."""
  @classmethod
  def wrapped_type(cls) -> Type[dict]:
    return dict

  @classmethod
  def set_wrapped_value(cls, target: dict, src: dict):
    target.clear()
    target.update(src)

@register
class ResourceTypeWrapper(FieldValueWrapper[type]):
  """Wrapper for type-valued fields."""

  @classmethod
  def wrapped_type(cls) -> Type[type]:
    return type



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the context module."""

import asyncio
import unittest
import threading
from typing import Any, Optional

from enact import contexts


@contexts.register
class SimpleContext(contexts.Context):
  """A context object for testing."""
  permissive_init: bool = True

  @classmethod
  def permissive_initialization(cls) -> bool:
    return cls.permissive_init

  def __init__(self):
    super().__init__()
    self.depth: Optional[int] = None

  def enter(self):
    assert self.depth is None
    cur = SimpleContext.get_current()
    if cur:
      assert cur.depth is not None
      self.depth = cur.depth + 1
    else:
      self.depth = 1

  def exit(self):
    self.depth = None


@contexts.register
class AsyncSimpleContext(contexts.AsyncContext):
  """An async context object for testing."""

  def __init__(self):
    super().__init__()
    self.depth: Optional[int] = None

  async def aenter(self):
    assert self.depth is None
    cur = AsyncSimpleContext.get_current()
    if cur:
      assert cur.depth is not None
      self.depth = cur.depth + 1
    else:
      self.depth = 1

  async def exit(self):
    self.depth = None


class UnregisteredContext(SimpleContext):
  """A context that was not registered."""


@contexts.register_to_superclass(SimpleContext)
class DerivedContextA(SimpleContext):
  """A derived context from SimpleContext."""


@contexts.register_to_superclass(SimpleContext)
class DerivedContextB(SimpleContext):
  """Another derived context from SimpleContext."""


class UnregisteredDerivedContext(SimpleContext):
  """A derived context that was not registered."""


class ContextTest(unittest.TestCase):
  """Test cases for the context class."""

  def test_current(self):
    """Tests that the current function works."""
    self.assertIs(SimpleContext.get_current(), None)
    with SimpleContext() as s1:
      self.assertEqual(SimpleContext.get_current(), s1)
      self.assertEqual(s1.depth, 1)
      with SimpleContext() as s2:
        self.assertEqual(SimpleContext.get_current(), s2)
        self.assertEqual(s2.depth, 2)
      self.assertEqual(SimpleContext.get_current(), s1)
    self.assertIs(SimpleContext.get_current(), None)

  def test_unregistered_fails(self):
    """Tests that using an unregistered context fails."""
    with self.assertRaises(contexts.ContextError):
      UnregisteredContext.get_current()

  def test_top_level(self):
    """Tests that top_level works as expected."""
    with SimpleContext() as s1:
      self.assertEqual(SimpleContext.current(), s1)
      with SimpleContext.top_level():
        self.assertEqual(SimpleContext.get_current(), None)
      self.assertEqual(SimpleContext.current(), s1)

  def test_derived_context(self):
    """Tests that derived contexts work as expected."""
    with DerivedContextA() as a:
      self.assertEqual(SimpleContext.get_current(), a)
      self.assertEqual(DerivedContextA.get_current(), a)

  def test_derived_context_different_subclass(self):
    """Tests that sibling derived contexts work as expected."""
    with DerivedContextA():
      with self.assertRaisesRegex(contexts.ContextError, 'not of type'):
        DerivedContextB.get_current()

  def test_derived_context_unregistered(self):
    """Tests that unregistered derived contexts fail."""
    with self.assertRaisesRegex(contexts.ContextError, 'not registered'):
      with UnregisteredDerivedContext():
        pass

  def test_register_to_superclass_unregistered_superclass(self):
    """Tests registering to an unregistered superclass."""
    class BaseContext(contexts.Context):
      pass

    class DerivedContext(BaseContext):
      pass

    with self.assertRaisesRegex(AssertionError, 'Superclass .* not registered'):
      contexts.register_to_superclass(BaseContext)(DerivedContext)

  def test_register_to_superclass_not_subclass(self):
    """Tests registering to a non-parent class."""
    @contexts.register
    class BaseContext(contexts.Context):
      pass

    class NonDerivedContext(contexts.Context):
      pass

    with self.assertRaisesRegex(AssertionError, 'must be a subclass'):
      contexts.register_to_superclass(BaseContext)(NonDerivedContext)

  def test_register_to_superclass_twice(self):
    """Tests registering to superclass twice."""
    @contexts.register
    class BaseContext(contexts.Context):
      pass

    class DerivedContext(BaseContext):
      pass

    contexts.register_to_superclass(BaseContext)(DerivedContext)
    with self.assertRaisesRegex(AssertionError, 'already registered'):
      contexts.register_to_superclass(BaseContext)(DerivedContext)

  def test_register_to_superclass_seperately_registered(self):
    """Tests registering to superclass after regular registration."""
    @contexts.register
    class BaseContext(contexts.Context):
      pass

    @contexts.register
    class DerivedContext(BaseContext):
      pass

    with self.assertRaisesRegex(AssertionError, 'already registered'):
      contexts.register_to_superclass(BaseContext)(DerivedContext)


class ContextDecoratorTest(unittest.TestCase):
  """Tests the context decorators."""

  def setUp(self):
    self.thread_result: Any = None
    self.context = SimpleContext()

  def enter_context(self):
    try:
      with self.context:
        self.thread_result = self.context.depth
    except contexts.ContextError as e:
      self.thread_result = e

  def test_threading_fails_if_not_permissive(self):
    """Tests that threading without a decorator fails."""
    SimpleContext.permissive_init = False
    t = threading.Thread(target=self.enter_context)
    t.start()
    t.join()
    self.assertIsInstance(self.thread_result, contexts.ContextError)

  def test_threading_succeeds_if_permissive(self):
    """Tests that threading without a decorator fails."""
    SimpleContext.permissive_init = True
    t = threading.Thread(target=self.enter_context)
    t.start()
    t.join()
    self.assertEqual(self.thread_result, 1)

  def test_with_new_contexts(self):
    """Tests that running with new contexts works."""
    with SimpleContext():
      t = threading.Thread(
        target=contexts.with_new_contexts(self.enter_context))
      t.start()
      t.join()
    self.assertEqual(self.thread_result, 1)

  def test_with_current_contexts(self):
    """Tests that running with the current contexts works."""
    with SimpleContext():
      t = threading.Thread(
        target=contexts.with_current_contexts(self.enter_context))
      t.start()
      t.join()
    self.assertEqual(self.thread_result, 2)

  def test_async_contexts(self):
    """Test async contexts."""
    async def _do_test():
      async with AsyncSimpleContext():
        self.assertEqual(AsyncSimpleContext.get_current().depth, 1)
        async with AsyncSimpleContext():
          self.assertEqual(AsyncSimpleContext.get_current().depth, 2)
        self.assertEqual(AsyncSimpleContext.get_current().depth, 1)

    asyncio.run(_do_test())



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Generic UI components."""

import abc
import json
from typing import Any, Callable, Dict, List, Optional, Tuple, Type, TypeVar, cast

import gradio as gr  # type: ignore
from gradio import events as gradio_events
import PIL.Image

from enact import contexts, resource_registry
from enact import interfaces
from enact import invocations
from enact import references
from enact import pretty_print
from enact import serialization


W = TypeVar('W', bound='EnactWidget')
C = TypeVar('C', bound=gr.components.Component)


_DEFAULT_INPUT_REQUIRED_TYPES = [
  str, type(None), int, float, bool, PIL.Image.Image]


class EnactWidget(abc.ABC):
  """A UI widget that represents an enact value."""

  def __init__(self):
    """Initializes the component."""
    self.components: List[gr.components.Component] = []

  @classmethod
  @abc.abstractmethod
  def supports_type(cls, value_type: Type) -> bool:
    """Returns whether the widget supports the value type."""

  @classmethod
  def create(cls: Type[W],
             value_type: Type,
             **kwargs) -> W:
    """Create widget for a value type or raises an error if not supported."""
    if not cls.supports_type(value_type):
      raise TypeError(f'{cls} does not support {value_type}')
    return cls(**kwargs)

  def add(self, c: C) -> C:
    """Adds a component to the widget."""
    self.components.append(c)
    return c

  @abc.abstractmethod
  def set(self,
          value: Optional[Any]=None,
          **kwargs) -> List[Dict[str, Any]]:
    """Set content and properties of UI elements to resource."""

  @abc.abstractmethod
  def get(self, *component_values) -> Optional[Any]:
    """Returns the current contents as a resource."""

  def consume(self, component_values: List) -> (
      Optional[interfaces.ResourceBase]):
    """Returns the contents as a resource and removes args from input list."""
    return self.get(*self.consume_args(component_values))

  def consume_args(self, component_values: List) -> List:
    """Return and consume the component arts."""
    component_args = component_values[:len(self.components)]
    del component_values[:len(self.components)]
    return component_args

  def set_from_event(
      self,
      event: gradio_events.EventListenerMethod,
      fun: Callable[..., Optional[Any]],
      inputs: List[gr.components.Component],
      **kwargs) -> gradio_events.Dependency:
    """Set from an event."""
    def _set(*args):
      value = fun(*args)
      updates = self.set(value, **kwargs)
      if len(updates) == 1:
        # Gradio needs unpacking for singleton lists.
        return updates[0]
      return updates
    return event(
      contexts.with_current_contexts(_set),
      inputs=inputs,
      outputs=self.components)


RW = TypeVar('RW', bound='RefWidget')


class RefWidget(EnactWidget):
  """A UI widget that represents a reference."""

  def _create_digest_display(self, **kwargs) -> gr.components.Component:
    """Returns a component displaying the digest of the reference."""
    return gr.Textbox(show_label=False, **kwargs)

  def _set_digest_display(self, value, **kwargs) -> gr.components.Component:
    """Returns a component displaying the digest of the reference."""
    display = cast(gr.Textbox, self.digest_display)
    return display.update(value=value, **kwargs)

  def __init__(self, **kwargs):
    super().__init__()
    with gr.Group():
      self.digest_display = self.add(self._create_digest_display(**kwargs))
      with gr.Accordion(label='Referenced resource', open=True):
        kwargs['interactive'] = False
        self.ref_details = self.add(
          gr.Markdown(**kwargs, value='```\nNo reference selected\n```'))
    self.change = self.digest_display.change

    def changed_ref_id(ref_id: str) -> Dict:
      """Performs necessary updates when ref id changes."""
      try:
        ref: references.Ref = references.Ref.from_id(ref_id)
      except json.JSONDecodeError:
        return self.ref_details.update(value='Invalid reference id.')
      try:
        resource = ref()
      except references.NotFound:
        return self.ref_details.update(value='Reference not found.')
      return self.ref_details.update(
        value=f'```\n{pretty_print.pformat(resource)}\n```')

    self.change(contexts.with_current_contexts(changed_ref_id),
                inputs=[self.digest_display],
                outputs=[self.ref_details])

  @classmethod
  def supports_type(cls, resource_type: Type) -> bool:
    """Returns whether the widget supports the resource type."""
    return issubclass(resource_type, references.Ref)

  def set(self,
          value: Optional[Any]=None,
          **kwargs) -> List[Dict[str, Any]]:
    """Set content and properties of UI elements."""
    if not value:
      return [
        self._set_digest_display(value='', **kwargs),
        self.ref_details.update(**kwargs)]
    assert isinstance(value, references.Ref)
    return [
        self._set_digest_display(value=value.id, **kwargs),
      self.ref_details.update(**kwargs)]

  def get(self, *component_values) -> Optional[interfaces.ResourceBase]:
    """Returns the current contents as a resource."""
    ref_box_value = component_values[0]
    if not ref_box_value:
      return None
    return references.Ref.from_id(ref_box_value)

class InvocationHistory(RefWidget):
  """A UI widget that represents a reference."""

  def __init__(self, **kwargs):
    super().__init__(**kwargs)
    self._state = gr.State(value=[])

    def _add_to_state(invocation_id: str, state_value: List[str]):
      try:
        references.Ref.from_id(invocation_id)
      except:  # pylint: disable=bare-except
        return state_value
      if invocation_id not in state_value:
        # We've already seen this invocation.
        state_value.append(invocation_id)
      return state_value, self.digest_display.update(choices=state_value)

    self.change(
      _add_to_state,
      inputs=[self.digest_display, self._state],
      outputs=[self._state, self.digest_display])

  def _create_digest_display(self, **kwargs) -> gr.components.Component:
    """Returns a component displaying the digest of the reference."""
    return gr.Dropdown(show_label=False, allow_custom_value=True, **kwargs)

  def _set_digest_display(self, value, **kwargs) -> gr.components.Component:
    """Returns a component displaying the digest of the reference."""
    display = cast(gr.Dropdown, self.digest_display)
    return display.update(value=value, **kwargs)


class ImageWidget(EnactWidget):
  """Supports image-type resources."""

  def __init__(self, **kwargs):
    super().__init__()
    self._image = self.add(gr.Image(**kwargs))

  @classmethod
  def supports_type(cls, value_type: Type) -> bool:
    """Supports all resources."""
    return issubclass(value_type, PIL.Image.Image)

  def set(self,
          value: Optional[Any]=None,
          **kwargs) -> List[Dict[str, Any]]:
    """Set content and properties of UI elements to resource."""
    assert not value or  isinstance(value, PIL.Image.Image)
    if value:
      return [self._image.update(value=value, **kwargs)]
    else:
      return [self._image.update(value=None, **kwargs)]

  def get(self, *component_values) -> Optional[Any]:
    """Returns the current contents as a resource."""
    if component_values[0] is None:
      return None
    return PIL.Image.fromarray(component_values[0])


class StrWidget(EnactWidget):
  """Supports string-type resources."""

  def __init__(self, **kwargs):
    super().__init__()
    self._box = self.add(gr.TextArea(**kwargs, show_label=False))

  @classmethod
  def supports_type(cls, value_type: Type) -> bool:
    """Supports all resources."""
    return issubclass(value_type, str)

  def set(self,
          value: Optional[Any]=None,
          **kwargs) -> List[Dict[str, Any]]:
    """Set content and properties of UI elements to resource."""
    if value is None:
      return [self._box.update(value='', **kwargs)]
    assert isinstance(value, str)
    return [self._box.update(value=str(value), **kwargs)]

  def get(self, *component_values) -> Optional[Any]:
    """Returns the current contents as a resource."""
    if component_values[0] is None:
      return None
    return str(component_values[0])


class JsonFieldWidget(EnactWidget):
  """Default widget to read / display a resource using JSON field input."""

  def __init__(self, resource_type: Type[interfaces.ResourceBase], **kwargs):
    """Initializes the component."""
    super().__init__()
    self._type = resource_type
    if not issubclass(self._type, interfaces.ResourceBase):
      self._type = resource_registry.wrap_type(self._type)

    self._serializer = serialization.JsonSerializer()
    self._boxes: Dict[str, gr.Textbox] = {}
    with gr.Group():
      for field_name in self._type.field_names():
        self._boxes[field_name] = self.add(gr.Textbox(
          label=field_name, **kwargs))

  @classmethod
  def supports_type(cls, value_type: Type) -> bool:
    """Supports all resources."""
    return True

  @classmethod
  def create(
      cls,
      value_type: Type,
      **kwargs) -> 'JsonFieldWidget':
    """Creates a new instance of the widget."""
    return cls(value_type, **kwargs)

  def set(self,
          value: Optional[Any]=None,
          **kwargs) -> List[Dict[str, Any]]:
    """Set content and properties of UI elements."""
    updates: List[Dict[str, Any]] = []
    if value:
      resource_dict = resource_registry.wrap(value).to_resource_dict()
    else:
      resource_dict = None

    for field_name in self._type.field_names():
      update_dict: Dict[str, Any] = {}
      if resource_dict:
        update_dict['value'] = json.dumps(
          self._serializer.to_json(resource_dict[field_name]),
          indent=2)
      else:
        update_dict['value'] = ''
      update_dict.update(kwargs)
      updates.append(self._boxes[field_name].update(**update_dict))
    return updates

  def get(self, *component_values) -> Any:
    """Returns the current contents as a resource."""
    resource_dict = interfaces.ResourceDict(self._type)
    for field_name, value in zip(self._type.field_names(), component_values):
      json_val = json.loads(value)
      resource_dict[field_name] = self._serializer.from_json(json_val)
    resource = resource_registry.from_resource_dict(resource_dict)
    return resource_registry.unwrap(resource)


class GUI:
  """A gradio GUI component for an invokable."""

  def __init__(
      self,
      invokable: references.Ref[invocations.InvokableBase],
      input_required_inputs:
        Optional[List[Type]]=None,
      input_required_outputs:
        Optional[List[Type]]=None):
    """Create a new UI for the component.

    Args:
      invokable: The invokable resource to create a UI for.
      input_required_inputs: Input types of invokables that may raise an
        input required exception.
      input_required_outputs: Output types of invokables that may raise an
        input required exception.
    """
    self._invokable = invokable
    self._input_required_inputs = cast(List[Type], list(set(
      _DEFAULT_INPUT_REQUIRED_TYPES).union(
        input_required_inputs or [])))
    self._input_required_outputs = cast(List[Type], list(set(
      _DEFAULT_INPUT_REQUIRED_TYPES).union(
        input_required_outputs or [])))

    input_type = self._invokable().get_input_type()
    if not input_type:
      raise ValueError('Input type must be specified.')
    self._input_type: Type[interfaces.ResourceBase] = input_type

    output_type = self._invokable().get_output_type()
    if not output_type:
      raise ValueError('Output type must be specified.')
    self._output_type: Type[interfaces.ResourceBase] = output_type

    self._blocks: Optional[gr.Blocks] = None
    self._input_widget: Optional[EnactWidget] = None
    self._output_widget: Optional[EnactWidget] = None
    self._invocation_widget: Optional[InvocationHistory] = None
    self._input_required_input_widgets: Dict[
      Type[interfaces.ResourceBase], EnactWidget] = {}
    self._input_required_output_widgets: Dict[
      Type[interfaces.ResourceBase], EnactWidget] = {}

    self._widget_types: List[Type[EnactWidget]] = []
    self.register(JsonFieldWidget)
    self.register(RefWidget)
    self.register(StrWidget)
    self.register(ImageWidget)

  @property
  def input_widget(self) -> EnactWidget:
    """Input widget of GUI."""
    assert self._input_widget, 'Blocks not generated yet.'
    return self._input_widget

  @property
  def output_widget(self) -> EnactWidget:
    """Output widget of GUI."""
    assert self._output_widget, 'Blocks not generated yet.'
    return self._output_widget

  @property
  def invocation_widget(self) -> InvocationHistory:
    """Invocation widget of GUI."""
    assert self._invocation_widget, 'Blocks not generated yet.'
    return self._invocation_widget

  def register(self, resource_widget: Type[EnactWidget]):
    """Register a resource widget."""
    self._widget_types.append(resource_widget)

  @property
  def blocks(self) -> gr.Blocks:
    """Blocks object representing the UI."""
    if not self._blocks:
      self._blocks = self._create_blocks()
    return self._blocks

  def _create_widget_by_resource_type(
      self,
      resource_type: Type[interfaces.ResourceBase],
      **kwargs) -> EnactWidget:
    """Create a widget for a resource type."""
    for widget_type in self._widget_types[::-1]:
      if widget_type.supports_type(resource_type):
        return widget_type.create(resource_type, **kwargs)
    raise TypeError(f'No widget for type {resource_type} found.')

  def _get_input_widget(self, resource_type: Type[interfaces.ResourceBase]):
    """Returns the widget for the input resource type."""
    return self._create_widget_by_resource_type(resource_type)

  def _get_output_widget(self, resource_type: Type[interfaces.ResourceBase]):
    """Returns the widget for the input resource type."""
    return self._create_widget_by_resource_type(
      resource_type, interactive=False)

  def _invoke(self, *args) -> Optional[references.Ref[invocations.Invocation]]:
    """Invoke the object."""
    input_value = self.input_widget.get(
      *args[:len(self.input_widget.components)])
    last_invocation = self.invocation_widget.get(*args[len(
      self.input_widget.components):])
    invokable = self._invokable
    if last_invocation:
      assert isinstance(last_invocation, references.Ref)
      invokable = last_invocation().response().invokable
    invocation = invokable().invoke(references.commit(input_value))
    return references.commit(invocation)

  def _continue(self, *args) -> Optional[
      references.Ref[invocations.Invocation]]:
    """Invoke the object."""
    component_values = list(args)
    invocation_ref = self.invocation_widget.consume(component_values)
    assert isinstance(invocation_ref, references.Ref)
    invocation = invocation_ref()
    assert isinstance(invocation, invocations.Invocation)
    raised = invocation.get_raised()
    assert isinstance(raised, invocations.InputRequest)
    requested_type = raised.requested_type

    user_input: Optional[interfaces.ResourceBase] = None
    for handled_type, widget in self._input_required_input_widgets.items():
      if handled_type == requested_type:
        user_input = widget.consume(component_values)
      else:
        widget.consume_args(component_values)
    continued = raised.continue_invocation(invocation, user_input)
    return references.commit(continued)

  def _title(self, ref: references.Ref) -> str:
    return f'### *{type(ref()).__name__}* `{ref.digest[:6]}`'

  def _create_blocks(self) -> gr.Blocks:
    """Return the gradio Blocks object representing the UI."""
    with gr.Blocks() as blocks:
      with gr.Group():
        title = gr.Markdown(value=self._title(self._invokable))
        with gr.Accordion(label='Resource details', open=False):
          pformat = pretty_print.pformat(
            self._invokable(), skip_repeated_refs=True)
          invokable_details = gr.Markdown(
            value=f'```{pformat}```')

      with gr.Group():
        gr.Markdown(value='Input:')
        self._input_widget = self._create_widget_by_resource_type(
          self._input_type)

      # Row workaround, see https://github.com/gradio-app/gradio/issues/4505.
      output_group = gr.Row(visible=False)
      with output_group, gr.Group():
        gr.Markdown(value='Output:')
        self._output_widget = self._create_widget_by_resource_type(
          self._output_type, interactive=False)

      # Row workaround, see https://github.com/gradio-app/gradio/issues/4505.
      exception_group = gr.Row(visible=False)
      with exception_group, gr.Group():
        gr.Markdown(value='Exception:')
        exception_area = gr.TextArea(
          value='', interactive=False, show_label=False)

      with gr.Group():
        input_required_header_md = gr.Markdown(visible=False)
        for resource_type in self._input_required_outputs:
          # Create widgets to display info to the user when querying for input.
          self._input_required_output_widgets[resource_type] = (
            self._create_widget_by_resource_type(
              resource_type, interactive=False, visible=False))
        input_required_input_md = gr.Markdown(visible=False)
        for resource_type in self._input_required_inputs:
          # Create widgets to sample input from the user.
          self._input_required_input_widgets[resource_type] = (
            self._create_widget_by_resource_type(
              resource_type, interactive=True, visible=False))
        continue_button = gr.Button(value='Continue', visible=False)

      with gr.Group():
        with gr.Accordion(label='Invocation details', open=False):
          self._invocation_widget = InvocationHistory()

      run_button = gr.Button('Run')
      def get_invocation(*args):
        """Grab the current invocation."""
        ref = self.invocation_widget.get(*args)
        if not ref:
          return None
        assert isinstance(ref, references.Ref)
        invocation = ref()
        assert isinstance(invocation, invocations.Invocation)
        return invocation

      def invocation_input(*args):
        """Return the input of the current invocation."""
        return get_invocation(*args).request().input()

      def invocation_output(*args):
        """Return the output of the current invocation."""
        output = get_invocation(*args).response().output
        if not output:
          return None
        return output()

      def flatten(l):
        """Flatten a list of lists."""
        return [item for sublist in l for item in sublist]

      def component_update(widget: EnactWidget, **kwargs):
        """Return a list of updates for the widget components."""
        return [c.update(**kwargs) for c in widget.components]

      def update_input_required_widgets(
          updates_list: List[Dict],
          input_widgets: bool,
          input_type: Optional[Type[interfaces.ResourceBase]]=None,
          input_resource: Optional[interfaces.ResourceBase]=None,
          **kwargs) -> bool:
        """Updates widgets that handle InputRequired exceptions."""
        found = False
        if input_widgets:
          input_required_widgets = self._input_required_input_widgets
        else:
          input_required_widgets = self._input_required_output_widgets

        for resource_type, widget in input_required_widgets.items():
          if input_type and issubclass(input_type, resource_type):
            found = True
            kwargs['visible'] = True
            updates_list += widget.set(input_resource, **kwargs)
          else:
            kwargs['visible'] = False
            updates_list += component_update(widget, **kwargs)
        return found

      def handle_invocation_exception(*args):
        """Enable / disable exception and output field."""
        invocation = get_invocation(*args)
        exception_ref = invocation.response().raised
        if not exception_ref:
          updates = [
            exception_group.update(visible=False),
            exception_area.update(value=''),
            output_group.update(visible=True),
            input_required_header_md.update(visible=False),
            input_required_input_md.update(visible=False),
            continue_button.update(visible=False)]
          update_input_required_widgets(
            updates, input_widgets=False, visible=False)
          update_input_required_widgets(
            updates, input_widgets=True, visible=False)
          return updates

        exception = exception_ref()
        if not isinstance(exception, invocations.InputRequest):
          updates = [
            exception_group.update(visible=True),
            exception_area.update(value=str(exception)),
            output_group.update(visible=False),
            input_required_header_md.update(visible=False),
            input_required_input_md.update(visible=False),
            continue_button.update(visible=False)]
          update_input_required_widgets(
            updates, input_widgets=False, visible=False)
          update_input_required_widgets(
            updates, input_widgets=True, visible=False)
          return updates

        # Try and display the exception causing input to the user.
        output_to_user = cast(interfaces.ResourceBase, exception.for_value())
        requested_input_from_user_type = exception.requested_type
        updates = []
        found_output_to_user_widget = update_input_required_widgets(
          updates, input_widgets=False,
          input_resource=output_to_user,
          input_type=type(output_to_user))
        found_input_from_user_widget = update_input_required_widgets(
          updates, input_widgets=True,
          input_type=requested_input_from_user_type)

        # If we found a way to display the output, use it, otherwise just dump
        # the exception as text.
        md_value = ''
        if not found_input_from_user_widget:
          md_value = (
            f'User input requested, but no widget found for '
            f'{exception.requested_type}')
        if exception.context:
          md_value += (
            f'\n\n*Context:*\n\n{pretty_print.pformat(exception.context)}')
        if not md_value:
          md_value = 'Provide input below.'

        updates_prefix = [
          exception_group.update(visible=not found_output_to_user_widget),
          exception_area.update(value='' if found_output_to_user_widget
                                else str(exception)),
          output_group.update(visible=False),
          input_required_header_md.update(
            value='### User input interrupt',
            visible=True),
          input_required_input_md.update(value=md_value, visible=True),
          continue_button.update(visible=found_input_from_user_widget)]
        return updates_prefix + updates

      def update_title(*args):
        response = get_invocation(*args).response
        assert response
        pformat = pretty_print.pformat(response().invokable,
                                       skip_repeated_refs=True)
        return (
          title.update(value=self._title(response().invokable)),
          invokable_details.update(
            value=f'```{pformat}```'))

      # Run the invokable on clicking Run.
      self.invocation_widget.set_from_event(
        run_button.click,
        contexts.with_current_contexts(self._invoke),
        inputs=self._input_widget.components +
               self._invocation_widget.components)

      # Continue the invocation on click Continue.
      self.invocation_widget.set_from_event(
        continue_button.click,
        contexts.with_current_contexts(self._continue),
        inputs=self._invocation_widget.components + flatten(
          [w.components for w in self._input_required_input_widgets.values()]
        ))

      # Set the input if the invocation changes.
      self._input_widget.set_from_event(
        self.invocation_widget.change,
        contexts.with_current_contexts(invocation_input),
        inputs=self.invocation_widget.components)

      # Set the output if the invocation changes.
      self._output_widget.set_from_event(
        self.invocation_widget.change,
        contexts.with_current_contexts(invocation_output),
        inputs=self.invocation_widget.components)

      # Handle exception.
      self._invocation_widget.change(
        contexts.with_current_contexts(handle_invocation_exception),
        inputs=self._invocation_widget.components,
        outputs=[
          exception_group, exception_area, output_group,
          input_required_header_md, input_required_input_md,
          continue_button] +
          flatten([w.components
                  for w in self._input_required_output_widgets.values()]) +
          flatten([w.components
                  for w in self._input_required_input_widgets.values()]))

      # Update the invokable title if the the invokable was updated.
      self._invocation_widget.change(
        contexts.with_current_contexts(update_title),
        inputs=self._invocation_widget.components,
        outputs=[title, invokable_details])

      return blocks

  def launch(self, *args, use_queue: bool=True, **kwargs) -> (
      Tuple[Any, str, Optional[str]]):
    """Launch the gradio UI, passing arguments to Blocks.launch.

    Args:
      use_queue: Specify whether queue should be used. The queue is required
        for executions exceeding 60 seconds.

    Returns:
      Tuple containing the FastAPI app object running the demo, local URL, and
      optional public URL if called with share=True.
    """
    if use_queue:
      return self.blocks.queue().launch(*args, **kwargs)
    return self.blocks.launch(*args, **kwargs)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the type_wrappers module."""

import unittest

import enact

from enact import type_wrappers
from enact import types


class TestTypeWrappers(unittest.TestCase):
  """Tests all the non-fieldvalue wrappers."""

  TYPES = [
    (type_wrappers.TupleWrapper, (1, 2, 3)),
    (type_wrappers.SetWrapper, {1, 2, 3}),
    (type_wrappers.TypeDescriptorWrapper, types.ResourceType(
       type_wrappers.SetWrapper.type_key())),
    (type_wrappers.ModuleWrapper, enact),
  ]

  def setUp(self):
    """Sets up the test."""
    self.store = enact.Store()

  def test_ref_deref(self):
    """Test storing and dereferencing a resource from a store."""
    with self.store:
      for resource_type, value in self.TYPES:
        with self.subTest(resource_type=resource_type):
          restored = self.store.commit(value).checkout()
          self.assertIsInstance(restored, type(value))
          self.assertEqual(restored, value)



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""A unified registration function."""

import inspect
from typing import Callable, Optional, Tuple, Type, TypeVar, Union, cast

from enact import function_wrappers
from enact import interfaces
from enact import invocations
from enact import resource_registry


Registerable = Union[
  Type[interfaces.ResourceBase],
  Callable]

RegisterableT = TypeVar('RegisterableT', bound=Registerable)


def _to_python_type(t) -> Optional[Type]:
  """Try to extract a python type from a type annotation."""
  if t is None:
    return type(None)
  if t is inspect.Signature.empty:
    return None
  if isinstance(t, type):
    return t
  if hasattr(t, '__origin__'):  # Handle generics
    if isinstance(t.__origin__, type):
      return t.__origin__
  return None


def _infer_types(invokable: Type[invocations._InvokableBase]) -> Tuple[
    Optional[Type], Optional[Type]]:
  """Infer types from the invokable."""
  signature = inspect.signature(invokable.call)
  output_type = _to_python_type(signature.return_annotation)
  params = signature.parameters
  if 'self' not in params:
    raise TypeError(
      'Invokable "call" function must have a "self" parameter.')
  if len(params) > 2:  # We count self here.
    raise TypeError(
      'Invokable "call" function must have at most one non-self parameter.')
  if len(params) == 1:
    input_type: Optional[Type] = type(None)
  else:
    (_, param) = params.values()
    if param.kind not in (param.POSITIONAL_ONLY,
                          param.POSITIONAL_OR_KEYWORD):
      raise TypeError(
        'Invokable "call" function must have a single '
        'non-self positional parameter.')
    input_type = _to_python_type(param.annotation)
  return input_type, output_type


def register(registerable: RegisterableT) -> RegisterableT:
  """Register a resource class, wrapper or python callable."""
  if (isinstance(registerable, type) and
      issubclass(registerable, invocations._InvokableBase)):  # pylint: disable=protected-access
    input_type, output_type = _infer_types(registerable)
    return cast(RegisterableT, invocations.typed_invokable(
      input_type, output_type)(registerable))
  elif (isinstance(registerable, type) and
      issubclass(registerable, interfaces.ResourceBase)):
    return cast(RegisterableT, resource_registry.register(registerable))
  elif isinstance(registerable, type):
    raise TypeError(
      f'Cannot register type {registerable}. '
       'Only types that inherit from Resource or Invokable can be registered. '
       'Consider using a TypeWrapper to register non-Resource types.')
  elif callable(registerable):
    return cast(RegisterableT, function_wrappers.register(registerable))
  else:
    raise TypeError(f'Cannot register {registerable}')



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for invocations."""

import asyncio
import dataclasses
import random
import tempfile
import time
from typing import Any, Optional, cast
import unittest
from unittest import mock

import enact
from enact import invocations

@enact.register
class ValueErrorResource(invocations.ExceptionResource):
  """A value error that is also a resource."""


@enact.typed_invokable(int, str)
@dataclasses.dataclass
class IntToStr(enact.Invokable):
  """An invokable that converts an int to a string."""
  salt: str = ''

  def call(self, value: int) -> str:
    return str(value) + self.salt


@enact.typed_invokable(int, str)
class WrongOutputType(enact.Invokable):

  def call(self, value: int) -> str:
    return value  # type: ignore


@enact.typed_invokable(int, int)
@dataclasses.dataclass
class AddOne(enact.Invokable):
  fail: bool = False
  def call(self, input_resource: int) -> int:
    if self.fail:
      raise ValueErrorResource('fail')
    return input_resource + 1


@enact.register
@dataclasses.dataclass
class Fail(enact.Invokable):
  def call(self, arg: Any) -> int:
    raise ValueErrorResource('fail')


@enact.typed_invokable(int, int)
@dataclasses.dataclass
class NestedFunction(enact.Invokable):
  """A nested function that repeatedly calls another invokable."""
  fun: enact.InvokableBase = AddOne()
  iter: int = 10
  fail_on: Optional[int] = None

  def call(self, input_resource: int) -> int:
    for i in range(self.iter):
      if i == self.fail_on:
        input_resource = Fail()(input_resource)
      else:
        input_resource = self.fun(input_resource)
    return input_resource


class InvocationsTest(unittest.TestCase):
  """Tests invocations."""

  def setUp(self):
    # pylint: disable=consider-using-with
    self.dir = tempfile.TemporaryDirectory()
    self.backend = enact.FileBackend(self.dir.name)
    self.store = enact.Store(self.backend)

  def tearDown(self):
    self.dir.cleanup()

  def test_typed_invokable(self):
    """"Test that the decorator works as expected."""
    fun = IntToStr('salt')
    self.assertEqual(fun.get_input_type(), int)
    self.assertEqual(fun.get_output_type(), str)
    output = fun.call(1)
    # Input and output types should not show up as fields.
    self.assertEqual(list(fun.field_names()), ['salt'])
    # The following line should typecheck correctly.
    self.assertEqual(output, '1salt')

  def test_typecheck_input(self):
    fun = IntToStr('salt')
    with self.assertRaises(TypeError):
      fun('1')

  def test_typecheck_output(self):
    fun = WrongOutputType()
    with self.assertRaises(TypeError):
      fun(1)

  def test_auto_input_args(self):
    self.assertEqual(
      IntToStr()(1), '1')

  def test_auto_input_kwargs(self):
    self.assertEqual(
      IntToStr()(1), '1')

  def test_native_exception_str(self):
    """Tests native exceptions string representation."""
    type_name = 'ValueError'
    str_representation = 'Bad thing happened.'
    traceback = 'foo: 23'
    exception = enact.NativeException(
      type_name, str_representation, traceback)
    self.assertEqual(
      str(exception), 'ValueError: Bad thing happened.')

  def test_native_exception_references(self):
    """Tests native exception commit/checkout."""
    type_name = 'ValueError'
    str_representation = 'Bad thing happened.'
    traceback = 'foo: 23'
    exception = enact.NativeException(
      type_name, str_representation, traceback)
    with self.store:
      ref = self.store.commit(exception)
      # pylint: disable=protected-access
      ref._clear_cache()
      checked_out = ref.checkout()
      self.assertEqual(
        checked_out.type_name, type_name)
      self.assertEqual(
        checked_out.str_representation, str_representation)
      self.assertEqual(
        checked_out.traceback, traceback)

  def test_invoke_simple(self):
    with self.store:
      fun = IntToStr('salt')
      invocation = fun.invoke(
        enact.commit(1))
      want: enact.Invocation = enact.Invocation(
        request=enact.commit(
          enact.Request(enact.commit(fun),
                        enact.commit(1))),
        response=enact.commit(
          enact.Response(
            invokable=enact.commit(fun),
            output=enact.commit('1salt'),
            raised=None,
            raised_here=False,
            children=[])))
    self.assertEqual(
      invocation,
      want)

  def test_call_nested(self):
    """Test calling a nested function."""
    self.assertEqual(NestedFunction()(1), 11)

  def test_invoke_nested(self):
    """Test invoking a nested function."""
    with self.store:
      invocation = NestedFunction().invoke(
        enact.commit(1))
      output = invocation.get_output()
      self.assertEqual(output, 11)
      self.assertEqual(len(list(invocation.get_children())), 10)
      for i, child in enumerate(invocation.get_children()):
        self.assertEqual(child.request().input(), i + 1)
        output = child.get_output()
        self.assertEqual(output, i + 2)

  def test_invoke_fail(self):
    with self.store:
      invocation = NestedFunction(fail_on=3).invoke(
        enact.commit(1), wrap_exceptions=True)
      self.assertFalse(invocation.successful())
      exception = invocation.get_raised()
      self.assertIsInstance(
        exception,
        enact.ExceptionResource)
      assert invocation.response().children
      self.assertEqual(len(invocation.response().children), 4)
      for i, child in enumerate(invocation.get_children()):
        if i < 3:
          self.assertEqual(child.request().input(), i + 1)
          output = child.get_output()
          self.assertEqual(output, i + 2)
        else:
          self.assertEqual(child.request().input(), 4)
          self.assertFalse(child.successful())
          exception = child.get_raised()
          self.assertIsInstance(
            exception,
            enact.ExceptionResource)

  def test_no_parent_on_invoke(self):
    """Tests that invocations are tracked in a fresh context."""
    with mock.patch.object(time, 'time_ns', return_value=0):
      with self.store as store:
        fun = IntToStr()
        with invocations.Builder(fun, store.commit(1)) as builder:
          # Enter an unrelated context.
          nested = NestedFunction()
          inner_invocation = nested.invoke(store.commit(1))
          # The invocation should not be tracked in the builder.
          # pylint: disable=protected-access
          self.assertEqual(builder._children, None)
        # Redo the invocation outside the builder.
        outer_invocation = nested.invoke(store.commit(1))
        # Ensure that executing in the misleading context made no difference.
        self.assertEqual(outer_invocation, inner_invocation)

  def test_meta_invoke(self):
    """Tests that meta-invocations are tracked correctly."""
    @dataclasses.dataclass
    class MetaInvoke(enact.Invokable):
      invokable: enact.InvokableBase
      def call(self, input_resource: enact.ResourceBase):
        return self.invokable.invoke(enact.commit(input_resource))

    with self.store as store:
      fun = NestedFunction()
      meta_invocation = MetaInvoke(fun).invoke(store.commit(1))
      self.assertEqual(
        len(list(meta_invocation.get_children())), 0)
      invocation = meta_invocation.get_output()
      assert isinstance(invocation, enact.Invocation)
      self.assertEqual(
        len(list(invocation.get_children())), 10)

  def test_wrapped_resource(self):
    """Tests that exceptions are tracked as wrapped when enabled."""
    class PythonErrorOnInvoke(enact.Invokable):
      def call(self, unused_input: enact.ResourceBase):
        raise ValueError('foo')
    with self.store as store:
      fun = PythonErrorOnInvoke()
      # Test with exception wrapping enabled.
      invocation = fun.invoke(store.commit(5), wrap_exceptions=True)
      self.assertIsInstance(invocation.get_raised(), enact.NativeException)
      # Test with exception wrapping disabled.
      with self.assertRaisesRegex(ValueError, 'foo'):
        _ = fun.invoke(store.commit(5), wrap_exceptions=False)

  def test_raise_native_error(self):
    """Tests that exceptions are raised in native format."""
    class PythonErrorOnInvoke(enact.Invokable):
      def call(self, unused_input: enact.ResourceBase):
        raise ValueErrorResource('foo')
    class ExpectValueError(enact.Invokable):
      def call(self, unused_input: enact.ResourceBase):
        try:
          PythonErrorOnInvoke()(3)
        except ValueErrorResource:
          return 'Got value error'
        raise enact.ExceptionResource('Expected ValueErrorResource')

    with self.store as store:
      fun = ExpectValueError()
      invocation = fun.invoke(store.commit(5))
      self.assertEqual(
        invocation.get_output(),
        'Got value error')

  def test_raised_here(self):
    """Tests that the raised_here field is set correctly."""
    class PythonErrorOnInvoke(enact.Invokable):
      def call(self, unused_input: enact.ResourceBase):
        raise ValueErrorResource('foo')

    @dataclasses.dataclass
    class SubCall(enact.Invokable):
      invokable: enact.Ref[enact.InvokableBase]
      def call(self, input_resource: enact.ResourceBase):
        self.invokable.checkout()(input_resource)

    with self.store as store:
      error_fun = PythonErrorOnInvoke()
      subcall_1 = SubCall(store.commit(error_fun))
      subcall_2 = SubCall(store.commit(subcall_1))

      invocation = subcall_2.invoke(store.commit(5))
      self.assertFalse(invocation.get_raised_here())
      (subinvocation,) = invocation.get_children()
      self.assertFalse(subinvocation.get_raised_here())
      (subsubinvocation,) = subinvocation.get_children()
      self.assertTrue(subsubinvocation.get_raised_here())

  def test_no_reraise_in_replay(self):
    """Tests that exceptions are replayed."""
    native_errors_raised = 0
    @enact.register
    class PythonErrorOnInvoke(enact.Invokable):
      def call(self, unused_input: enact.ResourceBase):
        nonlocal native_errors_raised
        native_errors_raised += 1
        raise ValueErrorResource('foo')

    @enact.register
    @dataclasses.dataclass
    class SubCall(enact.Invokable):
      invokable: enact.Ref[enact.InvokableBase]
      def call(self, input_resource: enact.ResourceBase):
        self.invokable.checkout()(input_resource)

    @enact.register
    @dataclasses.dataclass
    class MyResource(enact.Resource):
      value: int

    with self.store as store:
      error_fun = PythonErrorOnInvoke()
      subcall_1 = SubCall(store.commit(error_fun))
      subcall_2 = SubCall(store.commit(subcall_1))
      invocation = subcall_2.invoke(store.commit(MyResource(5)))
      self.assertEqual(native_errors_raised, 1)
      with self.assertRaises(ValueErrorResource):
        with invocations.ReplayContext(subinvocations=[
            enact.commit(invocation)]):
          subcall_2(MyResource(5))
      self.assertEqual(native_errors_raised, 2)


  def test_input_changed_error(self):
    @enact.register
    @dataclasses.dataclass
    class Changeable(enact.Resource):
      x: int = 0

    @enact.register
    class ChangesInput(enact.Invokable):
      def call(self, input_resource: Changeable):
        input_resource.x += 1
        return input_resource

    with self.assertRaises(enact.InputChanged):
      with self.store as store:
        ChangesInput().invoke(store.commit(Changeable()))

  def test_replay_call(self):
    """Test replaying a call."""
    fun = NestedFunction(fail_on=3)
    with self.store:
      invocation = fun.invoke(
        enact.commit(0))
      with enact.ReplayContext(
          subinvocations=[enact.commit(invocation)],
          exception_override=lambda x: enact.commit(100)):
        result = fun(0)
      self.assertEqual(result, 106)

  def test_replay_call_override_none(self):
    """Test replaying a call with a None override."""
    @enact.register
    def foo() -> int:
      enact.request_input(type(None))
      return 1

    with self.store:
      invocation = enact.invoke(foo)
      with enact.ReplayContext(
          subinvocations=[enact.commit(invocation)],
          exception_override=lambda x: enact.commit(None)):
        result = foo()
      self.assertEqual(result, 1)

  def test_replay_invocation(self):
    """Test replaying an invocation."""
    fun = NestedFunction()
    with self.store:
      invocation = fun.invoke(enact.commit(0))
      replay = invocation.replay()
      self.assertEqual(invocation, replay)

  def test_rewind_call(self):
    """Test rewinding a call."""
    leaf_calls = 0

    @enact.typed_invokable(type(None), int)
    class Leaf(enact.Invokable):
      def call(self):
        nonlocal leaf_calls
        leaf_calls += 1
        return 1

    @enact.typed_invokable(type(None), int)
    class Nested(enact.Invokable):
      def call(self):
        leaf = Leaf()
        return leaf() + leaf() + leaf()

    fun = Nested()
    with self.store:
      invocation = fun.invoke()
      self.assertEqual(leaf_calls, 3)
      invocation.replay()
      self.assertEqual(leaf_calls, 3)
      invocation = invocation.rewind(2)
      invocation.replay()
      self.assertEqual(leaf_calls, 5)

  def test_replay_modifies_invokable(self):
    @enact.register
    @dataclasses.dataclass
    class Counter(enact.Invokable):
      call_count: int = 0
      def call(self, input_value: int):
        self.call_count += 1
        return input_value

    with self.store:
      counter = Counter()
      fun = NestedFunction(counter, iter=10)
      invocation = fun.invoke(
        enact.commit(0))
      self.assertEqual(counter.call_count, 10)

      # Modify output to ensure we got a replay and
      # not a reexecution.
      with invocation.response.modify() as response:
        response.output = enact.commit(100)

      fun.fun = Counter()
      result = invocation.replay()
      self.assertEqual(result.get_output(), 100)
      self.assertEqual(counter.call_count, 10)

  def test_replay_partial(self):
    """Test replaying a partial invocation."""
    fun = NestedFunction(fail_on=3)
    with self.store:
      invocation = fun.invoke(
        enact.commit(0))
      with invocation.response.modify() as response:
        del response.children[4:]
        with response.children[-1].modify() as child:
          with child.response.modify() as child_response:
            assert child_response.raised
            child_response.raised = None
            child_response.output = enact.commit(100)
      with enact.ReplayContext(
          subinvocations=[enact.commit(invocation)]):
        result = fun(0)
      self.assertEqual(result, 106)

  def test_replay_call_on_mismatch_nonstrict(self):
    """Test non-strict replays are ignored if arguments don't match."""
    fun = NestedFunction(fail_on=3)
    with self.store:
      invocation = fun.invoke(
        enact.commit(0))
      with enact.ReplayContext(
          subinvocations=[enact.commit(invocation)],
          exception_override=lambda x: enact.commit(100), strict=False):
        with self.assertRaises(ValueErrorResource):
          # Exception override is not active since we're ignoring the
          # replay.
          fun(1)

  def test_replay_call_on_mismatch_strict(self):
    """Test strict replays raise."""
    fun = NestedFunction(fail_on=3)
    with self.store:
      invocation = fun.invoke(
        enact.commit(0))
      with enact.ReplayContext(
          subinvocations=[enact.commit(invocation)],
          exception_override=lambda x: enact.commit(100), strict=True):
        with self.assertRaises(enact.ReplayError):
          fun(1)

  def test_invoke_with_replay(self):
    """Test replays are ignored if arguments don't match."""
    fun = NestedFunction(fail_on=3)
    with self.store:
      invocation = fun.invoke(
        enact.commit(0))
      invocation = fun.invoke(
        enact.commit(0),
        replay_from=invocation,
        exception_override=lambda x: enact.commit(100))
      self.assertEqual(invocation.get_output(), 106)

  def test_request_input(self):
    """Tests the RequestInput invokable."""
    fun = NestedFunction(fun=enact.RequestInput(str), iter=2)
    with self.store:
      inputs = ['foo', 'bar', 'bish']
      invocation = fun.invoke(enact.commit(inputs[0]))
      for cur_input, next_input in zip(inputs[:-1], inputs[1:]):
        raised = invocation.get_raised()
        assert isinstance(raised, enact.InputRequest)
        self.assertEqual(
          cast(str, raised.for_value.checkout()), cur_input)
        invocation = raised.continue_invocation(
          invocation, next_input)
      self.assertEqual(invocation.get_output(), 'bish')
      child_outputs = [
        c.get_output() for c in invocation.get_children()]
      self.assertEqual(child_outputs, inputs[1:])

  def test_request_input_fun(self):
    """Tests the request_input function."""
    class MyInvokable(enact.Invokable):
      def call(self, value: int):
        return enact.request_input(int) + value
    with self.store:
      invocation = MyInvokable().invoke(enact.commit(3))
      raised = invocation.get_raised()
      assert isinstance(raised, enact.InputRequest)
      invocation = raised.continue_invocation(invocation, 5)
      self.assertEqual(invocation.get_output(), 8)

  def test_input_requested_outside_invocation(self):
    with self.assertRaises(enact.InputRequestOutsideInvocation):
      enact.RequestInput(str)('foo')

  def test_request_input_class(self):
    with self.store:
      fun = enact.RequestInput(str, 'Context')
      invocation = fun.invoke(enact.commit('foo'))
      raised = invocation.get_raised()
      assert isinstance(raised, enact.InputRequest)
      assert raised.context == 'Context'
      self.assertEqual(raised.requested_type, str)

  def test_empty_call_args(self):
    """Tests that empty call args are ok for None inputs."""
    @enact.typed_invokable(type(None), str)
    class TextInput(enact.Invokable):
      def call(self):
        return 'Foo'

    with self.store:
      fun = TextInput()
      invocation = fun.invoke()
      self.assertEqual(invocation.get_output(), 'Foo')


@enact.register
@dataclasses.dataclass
class AsyncWrapper(enact.AsyncInvokable):
  """An async wrapper for a sync invokable."""
  fun: enact._InvokableBase
  async def call(self, arg=None):
    fun = self.fun
    # pylint: disable=not-callable
    if isinstance(fun, enact.InvokableBase):
      return fun(arg)
    if isinstance(fun, enact.AsyncInvokable):
      return await fun(arg)


@enact.typed_invokable(float, float)
@dataclasses.dataclass
class Sleep(enact.AsyncInvokable):
  """Calls asyncio.sleep and returns actual time slept."""
  async def call(self, value: float) -> float:
    before = time.time()
    await asyncio.sleep(value)
    after = time.time()
    return float(after - before)


@enact.register
@dataclasses.dataclass
class AwaitInTask(enact.AsyncInvokable):
  """Launches another invokable in a task and possibly awaits it."""
  fun: enact.AsyncInvokable

  async def call(self, value: enact.ResourceBase) -> enact.ResourceBase:
    return await asyncio.get_running_loop().create_task(self.fun(value))


@enact.register
@dataclasses.dataclass
class RunawayTasks(enact.AsyncInvokable):
  """Invalid invokable that returns before all tasks are complete."""

  async def call(self, unused_value: enact.ResourceBase):
    sleep = Sleep()
    sleeps = [sleep(i * 0.1) for i in range(10)]
    await asyncio.wait(sleeps, timeout=0.2)

@enact.typed_invokable(type(None), int)
class AsyncRollDie(enact.AsyncInvokable):
  async def call(self):
    await asyncio.sleep(0.1 * random.random())
    return random.randint(1, 6)


@enact.typed_invokable(int, int)
class AsyncRollConcurrentDice(enact.AsyncInvokable):
  async def call(self, n: int):
    result = await asyncio.gather(
      *[AsyncRollDie()() for _ in range(n)])
    return sum(result)


@enact.typed_invokable(type(None), type(None))
class AsyncFail(enact.AsyncInvokable):
  async def call(self):
    await asyncio.sleep(0.01)
    raise ValueErrorResource('AsyncFail')


@enact.typed_invokable(type(None), list)
@dataclasses.dataclass
class Gather(enact.AsyncInvokable):
  """Calls asyncio.gather on a list of invokables."""
  invokables: list

  async def call(self):
    results = await asyncio.gather(
      *[i(None) for i in self.invokables],
      return_exceptions=True)
    for r in results:
      if isinstance(r, Exception):
        raise r
    return list(results)


class AsyncInvocationsTest(unittest.TestCase):
  """Tests invocations."""

  def setUp(self):
    # pylint: disable=consider-using-with
    self.dir = tempfile.TemporaryDirectory()
    self.backend = enact.FileBackend(self.dir.name)
    self.store = enact.Store(self.backend)

  def tearDown(self):
    self.dir.cleanup()

  def test_call(self):
    """Tests calling an async invokable directly."""
    fun = AsyncWrapper(IntToStr())
    result = asyncio.run(fun(3))
    self.assertEqual(result, '3')

  def test_invoke(self):
    """Tests basic invocations."""
    fun = AsyncWrapper(IntToStr())
    with self.store:
      result = asyncio.run(fun.invoke(enact.commit(3)))
      self.assertEqual(
        result,
        enact.Invocation(
          enact.commit(enact.Request(
            enact.commit(fun),
            enact.commit(3))),
          enact.commit(enact.Response(
            enact.commit(fun),
            enact.commit('3'),
            None,
            False,
            [
              enact.commit(enact.Invocation(
                enact.commit(enact.Request(
                  enact.commit(IntToStr()),
                  enact.commit(3))),
                enact.commit(enact.Response(
                  enact.commit(IntToStr()),
                  enact.commit('3'),
                  None,
                  False,
                  []))
              ))
            ]),
          ))
        )

  def test_invoke_nested_async(self):
    """Tests basic invocations od nested async invokables."""
    fun = AsyncWrapper(AsyncWrapper(IntToStr()))
    with self.store:
      invocation = asyncio.run(fun.invoke(
        enact.commit(3)))
      self.assertEqual(invocation.get_output(), '3')

  def test_invoke_async_with_task(self):
    """Tests that created tasks are correctly tracked."""
    fun = AwaitInTask(Sleep())
    with self.store:
      invocation = asyncio.run(fun.invoke(
        enact.commit(0.01)))
      self.assertGreater(invocation.get_output(), 0.0)

  def test_invoke_async_with_runaway_task(self):
    """Tests that runaway tasks raise an error."""
    fun = RunawayTasks()
    with self.store:
      with self.assertRaises(enact.IncompleteSubinvocationError):
        asyncio.run(fun.invoke())

  def test_simple_replay_async(self):
    fun = AsyncRollConcurrentDice()
    with self.store:
      invocation = asyncio.run(fun.invoke(enact.commit(20)))
      replay = asyncio.run(invocation.replay_async())
      self.assertEqual(invocation, replay)

  def test_wrapped_resource_async(self):
    """Tests that exceptions are tracked as wrapped when enabled."""
    class PythonErrorOnInvoke(enact.AsyncInvokable):
      async def call(self, unused_input: enact.ResourceBase):
        raise ValueError('foo')
    with self.store as store:
      fun = PythonErrorOnInvoke()
      # Test with exception wrapping enabled.
      invocation = asyncio.run(
        fun.invoke(store.commit(5), wrap_exceptions=True))
      self.assertIsInstance(invocation.get_raised(), enact.NativeException)
      # Test with exception wrapping disabled.
      with self.assertRaisesRegex(ValueError, 'foo'):
        _ = asyncio.run(fun.invoke(store.commit(5), wrap_exceptions=False))

  def test_replay_async(self):
    """Tests async replays."""
    fun = AsyncRollConcurrentDice()
    with self.store:
      invocation = asyncio.run(fun.invoke(enact.commit(20)))
      rolls = [c.get_output() for c in invocation.get_children()]
      self.assertEqual(len(rolls), 20)

      # Delete 10 outputs in the middle.
      with invocation.response.modify() as response:
        response.output = None
        for i in range(5, 15):
          with response.children[i].modify() as child:
            child.clear_output()

      invocation = asyncio.run(invocation.replay_async())
      rerolls = [c.get_output() for c in invocation.get_children()]
      self.assertEqual(len(rerolls), 20)
      # Check that the replays are the same as the original rolls.

  def test_replay_async_call(self):
    """Tests async replays using calls."""
    fun = AsyncRollConcurrentDice()
    with self.store:
      invocation = asyncio.run(fun.invoke(enact.commit(20)))
      with enact.ReplayContext([enact.commit(invocation)]):
        result = asyncio.run(fun(20))
      self.assertEqual(invocation.get_output(), result)

  def test_replay_async_multiple_exceptions(self):
    """Tests async replays that collect multiple exceptions."""
    fun = Gather([AsyncFail() for _ in range(10)])
    with self.store:
      invocation_cor = fun.invoke()
      invocation = asyncio.run(invocation_cor)
      exceptions = [c.get_raised() for c in invocation.get_children()]
      self.assertEqual(len(exceptions), 10)

  def test_continue_invocation_async(self):
    """Tests that input requests can be continued."""
    # Request ten integers.
    fun = Gather([AsyncWrapper(enact.RequestInput(int))
                  for _ in range(10)])
    with self.store:
      invocation = asyncio.run(fun.invoke(enact.commit(None)))
      input_request = invocation.get_raised()
      assert isinstance(input_request, enact.InputRequest)
      invocation = asyncio.run(input_request.continue_invocation_async(
        invocation, 2))
      self.assertEqual(
        invocation.get_output(),
        [2 for _ in range(10)])

  def test_continue_invocation_async_targeted(self):
    """Tests that input requests can be continued with different values."""
    fun = Gather([AsyncWrapper(enact.RequestInput(int, i))
                  for i in range(10)])
    with self.store:
      invocation = asyncio.run(fun.invoke(enact.commit(None)))
      self.assertIsInstance(invocation.get_raised(), enact.InputRequest)
      def exception_override(exception_ref):
        input_request = exception_ref()
        assert isinstance(input_request, enact.InputRequest)
        context = cast(int, input_request.context)
        return enact.commit(context + 1)
      invocation = asyncio.run(invocation.replay_async(exception_override))
      self.assertEqual(invocation.get_output(), list(range(1, 11)))



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for utils."""

import dataclasses
import unittest
from unittest import mock
from typing import Any, Type

import enact
from enact import utils

class UtilsTest(unittest.TestCase):
  """Tests for utils."""

  @enact.register
  @dataclasses.dataclass
  class _TestResource(enact.Resource):
    """A test resource."""
    foo: int

    def _get_foo(self):
      return self.foo

    @utils.cached_property
    def value(self):
      return self._get_foo()

  def test_cached_property(self):
    """Tests that the cached property works as expected."""
    test_resource = UtilsTest._TestResource(0)

    def mock_get_foo() -> mock.Mock:
      # pylint: disable=protected-access
      return mock.Mock(
        spec=test_resource._get_foo, side_effect=test_resource._get_foo)

    mock_getter = mock_get_foo()
    with mock.patch.object(test_resource, '_get_foo', mock_getter):
      self.assertEqual(test_resource.value, 0)
      self.assertTrue(mock_getter.called)

    mock_getter = mock_get_foo()
    with mock.patch.object(test_resource, '_get_foo', mock_getter):
      # Should not be called again as the digest of the resource hasn't changed.
      self.assertEqual(test_resource.value, 0)
      self.assertFalse(mock_getter.called)

    # This should change the digest.
    test_resource.foo = 1
    mock_getter = mock_get_foo()
    with mock.patch.object(test_resource, '_get_foo', mock_getter):
      # Should be called again as the digest of the resource has changed.
      self.assertEqual(test_resource.value, 1)
      self.assertTrue(mock_getter.called)

    # This should change the digest back to the original value, but the cache
    # will have been cleared, so the getter will be called again.
    test_resource.foo = 0
    mock_getter = mock_get_foo()
    with mock.patch.object(test_resource, '_get_foo', mock_getter):
      self.assertEqual(test_resource.value, 0)
      self.assertTrue(mock_getter.called)

  def test_walk_resource(self):
    """Tests walk_resource and walk_resource_dict works as expected."""
    @enact.register
    @dataclasses.dataclass
    class TestResource(enact.Resource):
      """A test resource."""
      value: Any

    @dataclasses.dataclass
    class Wrapped:
      """A wrapped resource"""
      value: Any

    @enact.register
    @dataclasses.dataclass
    class Wrapper(enact.TypeWrapper[Wrapped]):
      """A simple wrapper"""
      value: Any

      @classmethod
      def wrap(cls, value: Wrapped) -> 'Wrapper':
        return cls(value=value.value)

      @classmethod
      def wrapped_type(cls) -> Type[Wrapped]:
        return Wrapped

      def unwrap(self) -> Wrapped:
        return Wrapped(value=self.value)

    test_instance = TestResource(
      [TestResource(1),
       TestResource(2),
       [1, 2, 3, 4],
       Wrapped({'a': Wrapped(Wrapped)})]
    )

    result = list(
      utils.walk_resource(test_instance))

    dict_result = list(
      utils.walk_resource_dict(test_instance.to_resource_dict()))

    self.assertEqual(
      result,
      [
        test_instance,
        TestResource(1),
        TestResource(2),
        Wrapper({'a': Wrapped(Wrapped)}),
        Wrapper(Wrapped)
      ])
    self.assertEqual(
      [r.to_resource_dict() for r in result],
      dict_result)


if __name__ == '__main__':
  unittest.main()



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for the pretty printer."""

import unittest

import enact


class PrettyPrinterTest(unittest.TestCase):
  """Tests for the pretty printer."""

  def test_pformat_field_value(self):
    """Tests that printing field values works."""
    test_cases = [
      (None, 'None'),
      (True, 'True'),
      (False, 'False'),
      (42, '42'),
      (3.14, '3.14'),
      (b'hello', '<5 bytes>'),
      ('hello', '\'hello\''),
      ([1, 2, 3], '[\n  1\n  2\n  3]'),
      ({'a': 1, 'b': 2}, '{\n  "a": 1\n  "b": 2}')]
    for field_value, expected in test_cases:
      with self.subTest(field_value=field_value):
        formatted = enact.pformat(field_value)
        self.assertEqual(formatted, expected)

  def test_pformat_wrapped_resource(self):
    """Tests that printing wrapped values works."""
    formatted = enact.pformat((1, 2, 3))
    self.assertEqual(
      formatted,
'''TupleWrapper:
  value:
    [
      1
      2
      3]''')



# Copyright 2023 Agentic.AI Corporation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Version information for Enact.

This file should not be moved out of the root directory.
"""

from importlib import metadata
import logging
import os


# The name of the pip distribution.
DIST_NAME = 'enact'
# The root directory containing all python files associated with this
# distribution.
DIST_PYTHON_DIR = os.path.abspath(os.path.dirname(__file__))


try:
  __version__ = metadata.version(DIST_NAME)
except metadata.PackageNotFoundError:
  __version__ = '0.0.0'
  logging.warning(
    'Could not find installed version for %s, using %s',
    DIST_NAME, __version__)
